'' System Version 2.003.rr001
'' loaderInit Changes:
''
''   2006/09/24 - Corrected action table for i2cRead0Cmd
''   2006/10/02 - Improved assembly comments.  Changed i2cReset, i2cStop
''   2006/11/02 - Changed data setup time.  Changed i2cReset timing
''   2006/11/03 - Changed read/write method speed to 100KHz
''   2006/11/04 - Added checkPresence, writeWait, and computeTimes methods
''   2006/11/06 - Limited boot loading to 32K - 16 (OS uses last 16 bytes)
''                Added ioVerifyCmd and verifyEEPROM method
''   2006/11/09 - Modified boot and verify to use the minimum of the actual
''                length of the program stored in vbase ($0008) or the
''                specified byte count.
''                Note also that these routines don't know about option bits.
''   2006/11/10 - Now control block address is passed to start routine
''   2007/01/09 - Added SPI routines for SD card FAT access
''   2007/01/13 - Corrected checksum when ioBoot or ioVerify
''   2007/01/13 - Stores stack marker & clears VAR area on ioBootCmd
''   2007/02/06 - Stores stack marker & clears VAR area on ioSpiBoot
''   2007/02/13 - Changed the way verify mode was done, combined code
''   2007/02/20 - Added ioStopLdr option to stop the loader's cog
''   2007/02/22 - Corrected bootSDCard.  Needs start and initSDCard calls first.
''   2009/04/02 - Cluso99: Fix to force the SD card to tristate DO in spiDoStop
''   2010/03/28 - hairymnstr: Removed i2c and booting, added RAM access for DracBlade
''   2010/04/13 - hairymnstr: Re-wrote spin interface to match mb_small_spi for DracBlade
''   2010/05/05 - hairymnstr: Much of the PASM re-written to be compatible with mb_small_spi interface
''                            now also supports HCSD cards
''   2010/05/06 - hairymnstr: General tidy-up and cleaned out a few useless longs


CON
  clkfreqVal  = $0000                          ' Current CLKFREQ value stored here
  clksetVal   = $0004                          ' Current CLKSET value stored here
  ' possible card types
  type_MMC      = 1
  type_SD       = 2
  type_SDHC     = 3

  ' Error codes
  ERR_CARD_NOT_RESET            = -1
  ERR_3v3_NOT_SUPPORTED         = -2
  ERR_OCR_FAILED                = -3
  ERR_BLOCK_NOT_LONG_ALIGNED    = -4
  '...
  ' These errors are for the assembly engine...they are negated inside, and need to be <= 511
  ERR_ASM_NO_READ_TOKEN         = 100
  ERR_ASM_BLOCK_NOT_WRITTEN     = 101
  ERR_ASM_BAD_COMMAND           = 102                   ' negated in asm
  ERR_ASM_SPI_ERROR             = 103
  ' NOTE: errors -128 to -255 are reserved for reporting R1 response errors
  '...
  ERR_SPI_ENGINE_NOT_RUNNING    = -999
  ERR_CARD_BUSY_TIMEOUT          = -1000

  ' SDHC/SD/MMC command set for SPI
  CMD0    = $40+0        ' GO_IDLE_STATE
  CMD1    = $40+1        ' SEND_OP_COND (MMC)
  ACMD41  = $C0+41       ' SEND_OP_COND (SDC)
  CMD8    = $40+8        ' SEND_IF_COND
  CMD9    = $40+9        ' SEND_CSD
  CMD10   = $40+10       ' SEND_CID
  CMD12   = $40+12       ' STOP_TRANSMISSION
  CMD13   = $40+13       ' SEND_STATUS
  ACMD13  = $C0+13       ' SD_STATUS (SDC)
  CMD16   = $40+16       ' SET_BLOCKLEN
  CMD17   = $40+17       ' READ_SINGLE_BLOCK
  CMD18   = $40+18       ' READ_MULTIPLE_BLOCK
  CMD23   = $40+23       ' SET_BLOCK_COUNT (MMC)
  ACMD23  = $C0+23       ' SET_WR_BLK_ERASE_COUNT (SDC)
  CMD24   = $40+24       ' WRITE_BLOCK
  CMD25   = $40+25       ' WRITE_MULTIPLE_BLOCK
  CMD55   = $40+55       ' APP_CMD
  CMD58   = $40+58       ' READ_OCR
  CMD59   = $40+59       ' CRC_ON_OFF

  ' buffer size for my debug cmd log
  LOG_SIZE = 256<<1
  speed_buf = $170

  LEDMASK = %1000

  DELAY = 80             ' needs to be clock speed/1000000

DAT
'' I'm placing these variables in a DAT section to make this driver a singleton.
'' If for some reason you really need more than one driver (e.g. if you have more
'' than a single SD socket), move these back into VAR.
SPI_engine_cog          long 0
' these are used for interfacing with the assembly engine | temporary initialization usage
SPI_command             long 0  ' "t", "r", "w", 0 =>done, <0 => error          | unused
SPI_block_index         long 0  ' which 512-byte block to read/write            | cnt at init
SPI_buffer_address      long 0  ' where to get/put the data in Hub RAM          | unused
HighLatchVal            long 0

PUB start( basepin )
{{
  This is a compatibility wrapper, and requires that the pins be
  both consecutive, and in the order DO CLK DI CS.
}}
  return start_explicit( basepin, basepin+1, basepin+2, basepin+3 )

PUB readblock( block_index, buffer_address )
  if SPI_engine_cog == 0
    abort ERR_SPI_ENGINE_NOT_RUNNING
  if (buffer_address & 3)
    abort ERR_BLOCK_NOT_LONG_ALIGNED
  SPI_block_index := block_index
  SPI_buffer_address := buffer_address
  SPI_command := "r"
  repeat while SPI_command == "r"
  if SPI_command < 0
    abort SPI_command

PUB writeblock( block_index, buffer_address )
  if SPI_engine_cog == 0
    abort ERR_SPI_ENGINE_NOT_RUNNING
  if (buffer_address & 3)
    abort ERR_BLOCK_NOT_LONG_ALIGNED
  SPI_block_index := block_index
  SPI_buffer_address := buffer_address
  SPI_command := "w"
  repeat while SPI_command == "w"
  if SPI_command < 0
    abort SPI_command


'' -- DB ------------------------------------------------------
PUB writeram(hub_address, ram_address, count)
  if SPI_engine_cog == 0
    abort ERR_SPI_ENGINE_NOT_RUNNING
  SPI_block_index := ram_address
  SPI_buffer_address := ((count & $FFFF) << 16) + (hub_address & $FFFF)
  SPI_command := "W"
  repeat while SPI_command == "W"
  if SPI_command < 0
    abort SPI_command

PUB readram(hub_address, ram_address, count)
  if SPI_engine_cog == 0
    abort ERR_SPI_ENGINE_NOT_RUNNING
  SPI_block_index := ram_address
  SPI_buffer_address := ((count & $FFFF) << 16) + (hub_address & $FFFF)
  SPI_command := "R"
  repeat while SPI_command == "R"
  if SPI_command < 0
    abort SPI_command

#ifdef DracbladeProp
PUB sethighaddr(value)
  HighLatchVal := (HighLatchVal & $F0) + (value & 7)
  setha(HighLatchVal)

PUB setdebugport(value)
  HighLatchVal := (HighLatchVal & $0F) + ((value & $F) << 4)
  setha(HighLatchVal)

PUB setledon
  HighLatchVal |= LEDMASK
  setha(HighLatchVal)

PUB setledoff
  HighLatchVal &= (!LEDMASK)
  setha(HighLatchVal)

PRI setha(val)
  if SPI_engine_cog == 0
    abort ERR_SPI_ENGINE_NOT_RUNNING
  SPI_block_index := val
  SPI_buffer_address := 0
  SPI_command := "H"
  repeat while SPI_command == "H"
  if SPI_command < 0
    abort SPI_command
#endif
'' -- /DB -----------------------------------------------------
PUB get_seconds
  if SPI_engine_cog == 0
    abort ERR_SPI_ENGINE_NOT_RUNNING
  SPI_command := "t"
  repeat while SPI_command == "t"
  ' secods are in SPI_block_index, remainder is in SPI_buffer_address
  return SPI_block_index

PUB get_milliseconds : ms
  if SPI_engine_cog == 0
    abort ERR_SPI_ENGINE_NOT_RUNNING
  SPI_command := "t"
  repeat while SPI_command == "t"
  ' secods are in SPI_block_index, remainder is in SPI_buffer_address
  ms := SPI_block_index * 1000
  ms += SPI_buffer_address * 1000 / clkfreq

PUB start_explicit( DO, CLK, DI, CS ) : card_type | tmp, i
{{
  Do all of the card initialization in SPIN, then hand off the pin
  information to the assembly cog for hot SPI block R/W action!
}}
  ' Start from scratch
  stop
  ' clear my log buffer
  {
  bytefill( @log_cmd_resp, 0, LOG_SIZE+1 )
  dbg_ptr := @log_cmd_resp
  dbg_end := dbg_ptr + LOG_SIZE
  '}
  ' wait ~4 milliseconds
  waitcnt( 500 + (clkfreq>>8) + cnt )
  ' (start with cog variables, _BEFORE_ loading the cog)
  pinDO := DO
  maskDO := |< DO
  pinCLK := CLK
  maskCLK := |< CLK
  pinDI := DI
  maskDI := |< DI
  maskCS := |< CS
  adrShift := 9 ' block = 512 * index, and 512 = 1<<9
  ' pass the output pin mask via the command register
  maskAll := maskCS | maskCLK | maskDI
  dira |= maskAll
  ' get the card in a ready state: set DI and CS high, send => 74 clocks
  outa |= maskAll
  repeat 4096
    outa[CLK]~~
    outa[CLK]~
  ' time-hack
  SPI_block_index := cnt
  ' reset the card
  tmp~
  repeat i from 0 to 9
    if tmp <> 1
      tmp := send_cmd_slow( CMD0, 0, $95 )
      if (tmp & 4)
        ' the card said CMD0 ("go idle") was invalid, so we're possibly stuck in read or write mode
        if i & 1
          ' exit multiblock read mode
          repeat 4
            read_32_slow        ' these extra clocks are required for some MMC cards
          send_slow( $FD, 8 )   ' stop token
          read_32_slow
          repeat while read_slow <> $FF
        else
          ' exit multiblock read mode
          send_cmd_slow( CMD12, 0, $61 )
  if tmp <> 1
    ' the reset command failed!
    crash( ERR_CARD_NOT_RESET )
  ' Is this a SD type 2 card?
  if send_cmd_slow( CMD8, $1AA, $87 ) == 1
    ' Type2 SD, check to see if it's a SDHC card
    tmp := read_32_slow
  ' check the supported voltage
    if (tmp & $1FF) <> $1AA
      crash( ERR_3v3_NOT_SUPPORTED )
    ' try to initialize the type 2 card with the High Capacity bit
    repeat while send_cmd_slow( ACMD41, |<30, $77 )
    ' the card is initialized, let's read back the High Capacity bit
    if send_cmd_slow( CMD58, 0, $FD ) <> 0
      crash( ERR_OCR_FAILED )
    ' get back the data
    tmp := read_32_slow
    ' check the bit
    if tmp & |<30
      card_type := type_SDHC
      adrShift := 0
    else
      card_type := type_SD
  else
    ' Either a type 1 SD card, or it's MMC, try SD 1st
    if send_cmd_slow( ACMD41, 0, $E5 ) < 2
      ' this is a type 1 SD card (1 means busy, 0 means done initializing)
      card_type := type_SD
      repeat while send_cmd_slow( ACMD41, 0, $E5 )
    else
      ' mark that it's MMC, and try to initialize
      card_type := type_MMC
      repeat while send_cmd_slow( CMD1, 0, $F9 )
    ' some SD or MMC cards may have the wrong block size, set it here
    send_cmd_slow( CMD16, 512, $15 )
  ' card is mounted, make sure the CRC is turned off
  send_cmd_slow( CMD59, 0, $91 )
  '  check the status
  'send_cmd_slow( CMD13, 0, $0D )
  ' done with the SPI bus for now
  outa |= maskCS
  ' Hand off control to the assembly engine's cog
  bufAdr := @SPI_buffer_address
  sdAdr := @SPI_block_index
  SPI_command := -1 ' just make sure it's not 1
  ' start my driver cog and wait till I hear back that it's done
  SPI_engine_cog := cognew( @SPI_engine_entry, @SPI_command ) + 1
  if( SPI_engine_cog == 0 )
    crash( ERR_SPI_ENGINE_NOT_RUNNING )
  repeat while SPI_command <> 0
  ' and we no longer need to control any pins from here
  dira &= !maskAll
  ' the return variable is card_type

PUB release
{{
  I do not want to abort if the cog is not
  running, as this is called from stop, which
  is called from start/ [8^)
}}
  if SPI_engine_cog
    SPI_command := "z"
    repeat while SPI_command == "z"

PUB stop
{{
  kill the assembly driver cog.
}}
  release
  if SPI_engine_cog
    cogstop( SPI_engine_cog~ - 1 )

PUB get_SPI_command
  return @SPI_command

PUB get_memory
  return @SPI_engine_entry

PRI crash( abort_code )
{{
  In case of Bad Things(TM) happening,
  exit as gracefully as possible.
}}
  ' and we no longer need to control any pins from here
  dira &= !maskAll
  ' and report our error
  abort abort_code

PRI send_cmd_slow( cmd, val, crc ) : reply | time_stamp
{{
  Send down a command and return the reply.
  Note: slow is an understatement!
  Note: this uses the assembly DAT variables for pin IDs,
  which means that if you run this multiple times (say for
  multiple SD cards), these values will change for each one.
  But this is OK as all of these functions will be called
  during the initialization only, before the PASM engine is
  running.
}}
  ' if this is an application specific command, handle it
  if (cmd & $80)
    ' ACMD<n> is the command sequense of CMD55-CMD<n>
      cmd &= $7F
      reply := send_cmd_slow( CMD55, 0, $65 )
      if (reply > 1)
        return reply
  ' the CS line needs to go low during this operation
  outa |= maskCS
  outa &= !maskCS
  ' give the card a few cocks to finish whatever it was doing
  read_32_slow
  ' send the command byte
  send_slow( cmd, 8 )
  ' send the value long
  send_slow( val, 32 )
  ' send the CRC byte
  send_slow( crc, 8 )
  ' is this a CMD12?, if so, stuff byte
  if cmd == CMD12
    read_slow
  ' read back the response (spec declares 1-8 reads max for SD, MMC is 0-8)
  time_stamp := 9
  repeat
    reply := read_slow
  while( reply & $80 ) and ( time_stamp-- )
  ' done, and 'reply' is already pre-loaded

PRI send_slow( value, bits_to_send )
  value ><= bits_to_send
  repeat bits_to_send
    outa[pinCLK]~
    outa[pinDI] := value
    value >>= 1
    outa[pinCLK]~~

PRI read_32_slow : r
  repeat 4
    r <<= 8
    r |= read_slow

PRI read_slow : r
{{
  Read back 8 bits from the card
}}
  ' we need the DI line high so a read can occur
  outa[pinDI]~~
  ' get 8 bits (remember, r is initialized to 0 by SPIN)
  repeat 8
    outa[pinCLK]~
    outa[pinCLK]~~
    r += r + ina[pinDO]
  ' error check
  if( (cnt - SPI_block_index) > (clkfreq << 2) )
    crash( ERR_CARD_BUSY_TIMEOUT )

DAT
                        org     0
SPI_engine_entry        or      outa, maskDI
                        or      outa, maskCLK
                        mov     dira, maskALL
done                    wrlong  err, PAR             ' status  =0=false=good, else error x

                        ' wait for a command (pause short time to reduce power)
pause                   mov     ctr, #DELAY      wz      ' if =0 no pause
              if_nz     add     ctr, cnt
              if_nz     waitcnt ctr, #0                 ' wait for a short time (reduces power)
                        rdlong  i2cCmd, PAR     wz      ' command ?
              if_z      jmp     #pause                  ' not yet
                        test    i2cCmd, #$100   wz
              if_nz     jmp     #pause

                        rdlong  i2cAddr, sdAdr
                        rdlong  i2cBufAdr, bufAdr
                        rdlong  SaveClkFreq,#clkfreqVal ' Save clock frequency and mode
                       ' rdbyte  SaveClkMode,#clksetVal
                        test    i2cCmd, #%100000      wz ' bit set  for lower case
                if_nz   jmp     #spiEntryPoint          ' if it's not an SPI command do RAM

ramEntryPoint
                        ' need to separate count and hub address
                        mov     i2cCount, i2cBufAdr
                        shr     i2cCount, #16
                        and     i2cBufAdr, i2cWordMask
                        cmp     i2cCmd, #"R"       wz      ' R = read block
              if_z      jmp     #rdblock
                        cmp     i2cCmd, #"W"       wz      ' W = write block
              if_z      jmp     #wrblock
                        cmp     i2cCmd, #"H"       wz      ' H sets the high latch
              if_z      jmp     #sethighlatch
                        neg     err, #ERR_ASM_BAD_COMMAND  ' error = cmd (unknown command)
                        jmp     #done

'' SPI routines for Rokicki's SD card FAT file system driver

spiEntryPoint           shl     i2cAddr, adrShift       ' byte to block for old SD does nothing for HCSD
                        cmp     i2cCmd,#"z" wz
              if_z      jmp     #spiDoStop
                        cmp     i2cCmd,#"r" wz
              if_z      jmp     #spiDoRead
                        cmp     i2cCmd,#"w" wz
              if_z      jmp     #spiDoWrite
                        neg     err, #ERR_ASM_BAD_COMMAND
                        jmp     #done

'' Stop SPI communications.  Any previously used I/O pins are set to input
'' mode and the masks for the I/O pins are zeroed.
' Modification by Cluso99 to force the SD card to tristate the D0 pin (allows other chips on the bus)
'' All we do here is set the CS high, clock the card a bit to make sure it releases DO and tri-state
'' the other pins.  Any card access routine sets pins as output again and set the CS etc.
spiDoStop               or      outa,maskCS          ' CS=1
                        andn    dira,maskDO          ' disable D0 as output (s/be input anyway)
                        call    #spiRecvByte            ' Output a stream of clocks
                        call    #spiRecvByte
                        andn    dira,maskCS          ' CS input
                        andn    dira,maskDI          ' DI as input
                        andn    dira,maskClk         ' CLK input
                        jmp     #i2cGoUpdate

'' Read one or more 512 byte blocks and store the specified number of bytes
'' into the HUB location given.  The block number is provided in the 24 bit
'' address field and incremented after every block is read.  Partial blocks are
'' allowed and any extra bytes read are discarded.

spiDoRead               or      dira, maskALL           ' make sure output pins are output
                        mov     spiOp,#17               ' READ_SINGLE_BLOCK
                        mov     i2cTime,cnt             ' Save start of timeout
                        mov     spiParm,i2cAddr
                        call    #spiSendCmd             ' Read from specified block
                        call    #spiResponse
                        mov     spiBlkCnt,spiBlkSize    ' Transfer a block at a time
:getRead                call    #spiRecvByte
                        call    #StoreData
                        djnz    spiBlkCnt,#:getRead     ' Are we done with the block?
                        call    #spiRecvByte
                        call    #spiRecvByte            ' Yes, finish with 16 clocks
                        jmp     #i2cGoUpdate            ' If not booting, we're done

'' Write one or more 512 byte blocks with the specified number of bytes from
'' the HUB location given.  The block number is provided in the 24 bit address
'' field and incremented after every block is written.  Partial blocks are
'' allowed and are padded with zeroes.

spiDoWrite              or      dira, maskALL           ' set pin directions
                        mov     spiOp,#24               ' WRITE_BLOCK
                        mov     i2cTime,cnt             ' Setup timeout
                        mov     spiParm,i2cAddr
                        call    #spiSendCmd             ' Write to specified block
                        mov     i2cData,#$FE            ' Ask to start data transfer
                        call    #spiSendByte
                        mov     spiBlkCnt,spiBlkSize    ' Transfer a block at a time
:putWrite               rdbyte  i2cData,i2cBufAdr       ' If not, get the next data byte
                        add     i2cBufAdr,#1
:padWrite               call    #spiSendByte
                        djnz    spiBlkCnt,#:putWrite    ' Are we done with the block?
                        call    #spiRecvByte
                        call    #spiRecvByte            ' Yes, finish with 16 clocks
                        call    #spiResponse
                        and     i2cData,#$1F            ' Check the response status
                        cmp     i2cData,#5         wz
              if_ne     neg     err, #ERR_ASM_SPI_ERROR ' Must be Data Accepted
              if_ne     jmp     #done
                        movs    spiWaitData,#0          ' Wait until not busy
                        call    #spiWaitBusy
                        jmp     #i2cGoUpdate

'' Mid level SPI I/O

spiSendCmd              andn    outa,maskCS          ' Send command sequence.  Begin by
                        call    #spiRecvByte            '  selecting card and clocking
                        mov     i2cData,spiOp
                        or      i2cData,#$40            ' Send command byte (1st 2 bits %01)
                        call    #spiSendByte
                        mov     i2cData,spiParm
                        shr     i2cData,#24             ' Supplied address is sector number
                        call    #spiSendByte 
                        mov     i2cData,spiParm         ' Send to SD card as byte address,
                        shr     i2cData,#16              '  in multiples of 512 bytes
                        call    #spiSendByte
                        mov     i2cData,spiParm         ' Total length of this address is
                        shr     i2cData,#8              '  four bytes
                        call    #spiSendByte
                        mov     i2cData,spiParm
                        call    #spiSendByte
                        mov     i2cData,#$95            ' CRC code (for 1st command only)
                        call    #spiSendByte
spiResponse             movs    spiWaitData,#$FF        ' Wait for response from card
spiWaitBusy             call    #spiRecvByte
                        mov     i2cTemp,cnt
                        sub     i2cTemp,i2cTime         ' Check for expired timeout (1 sec)
                        cmp     i2cTemp,SaveClkFreq wc
              if_nc     neg     err, #ERR_ASM_SPI_ERROR
              if_nc     jmp     #done
spiWaitData             cmp     i2cData,#0-0       wz   ' Wait for some other response
              if_e      jmp     #spiWaitBusy            '  than that specified
spiSendCmd_ret
spiResponse_ret
spiWaitBusy_ret         ret

'' Low level byte I/O

spiSendByte             mov     i2cMask,#%10000000
:sendBit                test    i2cMask,i2cData    wc
                        andn    outa,maskClk         ' Send data bytes MSB first
                        muxc    outa,maskDI
                        or      outa,maskClk
                        shr     i2cMask,#1              ' When mask shifted out, we're done
                        tjnz    i2cMask,#:sendBit
                        or      outa,maskDI          ' Leave DI in idle (high) state
spiSendByte_ret         ret

spiRecvByte             mov     i2cMask,#%10000000
:recvBit                andn    outa,maskClk         ' Receive data bytes MSB first
                        or      outa,maskClk         ' Copy DO to data bit
                        test    maskDO,ina      wc
                        muxc    i2cData,i2cMask
                        shr     i2cMask,#1              ' When mask shifted out, we're done
                        tjnz    i2cMask,#:recvBit
                        and     i2cData,#%11111111      ' Eight bits received
spiRecvByte_ret         ret


'' Just writes the byte to hub and then increments the pointer
StoreData               wrbyte  i2cData,i2cBufAdr       ' Store data in specified location
                        add     i2cBufAdr,#1            '  and increment the address
StoreData_ret           ret

'' Successful exit path, sets error to zero and then jumps back to write back
i2cGoUpdate             mov     err, #0
                        jmp     #done          '  and go wait for a new one

'=========================================================================================================
' RAM functions
'=========================================================================================================

'' set high address bytes with command H, pass value in third variable of the DoCmd
'' 4 bytes - masks off all but bits 16 to 23
sethighlatch            and     i2cAddr,#$FF               ' ensure rest of bits zero
                        or      dira,latchdirection        ' setup active pins 138 and bus
                        andn    outa, latchdirection
                        or      outa, i2cAddr              ' send out HighLatch
                        or      outa, HighAddress          ' or with the high address
                        andn    outa, GateHigh             ' set gate low
                        or      outa, GateHigh             ' set the gate high again
                        andn    dira, LatchDirection
                        mov     err, #0
                        jmp     #done

'Memory Access Functions
rdblock
rdloop                  call    #RamAddress                ' sets up the latches with the correct ram address
                        ' need to turn bus around after writing address
                        andn    dira, #$ff
                        andn    outa, LatchDirection       ' clear address bits (read = 0000 on 138)
                        andn    outa,GateHigh              ' set gate low
                        nop                                ' short delay to stabilise
                        nop
                        mov     data_8, ina                ' read SRAM
                        and     data_8, #$FF               ' extract 8 bits
                        or      outa,GateHigh              ' set the gate high again
                        ' now write the read byte to hub ram
                        wrbyte  data_8,i2cBufAdr           ' write data_8 to hubaddr ie copy byte to hub
                        add     i2cBufAdr,#1               ' add 1 to hub address
                        add     i2cAddr,#1                 ' add 1 to ram address
                        djnz    i2cCount,#rdloop           ' loop until done
                        ' all done,
                        mov     err, #0
                        andn    dira, LatchDirection
                        jmp     #done                      ' set error=False

wrblock
wrloop                  rdbyte  data_8, i2cBufAdr          ' copy byte from hub
                        call    #RamAddress                ' sets up the latches with the correct ram address
                        mov     outx,data_8                ' get the byte to output
                        and     outx, #$FF                 ' ensure upper bytes=0
                        or      outx,WriteEnable           ' or with correct 138 address
                        andn    outa,latchdirection
                        or      outa,outx                  ' send it out
                        andn    outa,GateHigh              ' set gate low
                        nop                                ' no nop doesn't work, one does, so put in two to be sure
                        nop                                ' another NOP
                        or      outa,GateHigh              ' set it high again
                        ' Now increment the pointers
                        add     i2cBufAdr,#1               ' add 1 to hub address
                        add     i2cAddr,#1                 ' add 1 to ram address
                        djnz    i2cCount,#wrloop           ' loop until done
                        ' Set done flag and go back to idle
                        mov     err, #0
                        andn    dira, LatchDirection
                        jmp     #done                      ' reinitialise

RamAddress              ' sets up the low 2 ram latches.
                        ' called for every byte.  To set the high latch send an "H" command
                        or dira,LatchDirection             ' set up the pins for programming latch chips
                        mov outx,i2cAddr                   ' get the address into a temp variable
                        and outx,#$FF                      ' mask the low byte
                        or  outx,LowAddress                ' or with 138 low address
                        andn outa, LatchDirection
                        or outa,outx                       ' send it out
                        andn outa,GateHigh                 ' set gate low
                        or outa,GateHigh                   ' set it high again
                        ' now repeat for the middle byte
                        mov outx,i2cAddr                   ' get the address into a temp variable
                        shr outx,#8                        ' shift right by 8 places
                        and outx,#$FF                      ' mask the low byte
                        or  outx,MiddleAddress             ' or with 138 middle address
                        andn outa, LatchDirection
                        or outa,outx                       ' send it out
                        andn outa,GateHigh                 ' set gate low
                        or outa,GateHigh                   ' set it high again
RamAddress_ret          ret


'delay                   long    80                                    ' waitcnt delay to reduce power (#80 = 1uS approx)
ctr                     long    0                                     ' used to pause execution (lower power use) & byte counter
GateHigh                long    %00000000_00000000_00000001_00000000  ' HC138 gate high, all others must be low
Outx                    long    0                                     ' for temp use, same as n in the spin code
LatchDirection          long    %00000000_00000000_00001111_11111111 ' 138 active, gate active and 8 data lines active
LowAddress              long    %00000000_00000000_00000101_00000000 ' low address latch = xxxx010x and gate high xxxxxxx1
MiddleAddress           long    %00000000_00000000_00000111_00000000 ' middle address latch = xxxx011x and gate high xxxxxxx1
HighAddress             long    %00000000_00000000_00001001_00000000 ' high address latch = xxxx100x and gate high xxxxxxx1
WriteEnable             long    %00000000_00000000_00000011_00000000 ' /WE = xxxx001x and gate high xxxxxxx1
data_8                  long    %00000000_00000000_00000000_00000000 ' so code compatability with zicog driver

'' Constants for all routines

i2cWordMask             long    $0000FFFF
spiBlkSize                                              ' Number of bytes in an SD card block
incrDst                 long    %10_00000000            ' Used to increment destination field
testIns                 shl     i2cData,#32             ' Used to compare for end of word packing

'' Variables for all routines

i2cCmd                  long    0
i2cTemp                 long    0
i2cCount                long    0
i2cBufAdr               long    0
i2cAddr                 long    0
SaveClkFreq             long    0                       ' Initial clock frequency (clkfreqVal)
spiBlkCnt               long    0                       ' Number of SD card bytes to go in block

'' Local variables for low level I2C routines

spiOp                   long    0                       ' Used as a temp for the SPI command
spiParm                 long    0                       ' Used for shifting the address out
i2cTime                 long    0                       ' Used for timekeeping
i2cData                 long    0                       ' Data to be transmitted / received
i2cMask                 long    0                       ' Bit mask for bit to be tx / rx
i2cBitCnt               long    0                       ' Number of bits to tx / rx

'' Additional local variables for SPI SD Card access

pinDO                   long    0
pinClk                  long    0
pinDI                   long    0
pinCS                   long    0
maskDO                  long    0
maskDI                  long    0
maskCLK                 long    0
maskCS                  long    0
maskALL                 long    0

bufAdr                  long    0
sdAdr                   long    0
err                     long    0
adrShift                long    0
                        fit     $170
