CON
	' 6502 flag registers
	CF			=	%00000001	' carry
	ZF			=	%00000010	' zero
	II			=	%00000100	' interrupt inhibit
	DF			=	%00001000	' decimal mode
	BF			=	%00010000	' break flag
	TF			=	%00100000	' trap flag
	VF			=	%01000000	' overflow
	NF			=	%10000000	' negative


	io_cmd_out		=	$01
	io_cmd_in		=	$02
	io_cmd_break		=	$03

	lmm_brk_		=	4* 0
	lmm_trace_		=	4* 1
	lmm_ill_		=	4* 2
	lmm_adc_		=	4* 3
	lmm_sbc_		=	4* 4
	lmm_jsr_		=	4* 5
	lmm_rts_		=	4* 6
	lmm_rti_		=	4* 7
	lmm_jmp_ind_		=	4* 8
	lmm_rra_		=	4* 9
	lmm_isb_		=	4*10
	lmm_rla_		=	4*11
	lmm_ssh_aby_np_		=	4*12
	lmm_syh_abx_np_		=	4*13
	lmm_sxh_aby_np_		=	4*14
	lmm_sah_idy_np_		=	4*15
	lmm_sah_aby_np_		=	4*16
	lmm_ast_aby_p_		=	4*17
	lmm_arr_imm_		=	4*18
	lmm_asx_imm_		=	4*19
	lmm_axa_imm_		=	4*20
	lmm_oal_imm_		=	4*21
{
'<><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>
CON           'THESE LINES ARE REQUIRED IN THE TOP OBJECT AND AT THE TOP OF THE CODE !!!
'------------------------------------------------------------------------------------------------------------
' The following is the object offset needed to be added to #@xxxx in Debug_Block pasm instructions (compiler restriction)
  DB            = $10                                          '<=====
 ' The following is the debug code which is held in shadow ram (compiler restriction)
  X_ENTER       = $1F0 ' X_ENTER   RDLONG  X_OPCODE, #D_LMM    'read instruction to be executed 
  X_ADD4        = $1F1 '           ADD     X_ENTER,  #4        'inc to next hub instruction to be executed
  X_OPCODE      = $1F2 ' X_OPCODE  NOP                         'execute the instruction ("SUB X_ENTER,#4 if waiting)
  X_1F3         = $1F3 '           JMP     #X_ENTER            'loop again
DAT           'THESE LINES ARE REQUIRED IN THE TOP OBJECT AND AT THE TOP OF THE CODE !!! 
'------------------------------------------------------------------------------------------------------------
' THIS IS THE DEBUG BLOCK WHICH MUST BE LOCATED IN LOWER HUB RAM (entirely below $200)
' In order to achieve this the DAT block must be at the top of the top object.
'------------------------------------------------------------------------------------------------------------
' Cog execution begins as follows:
'   The bootcode placed into cog $000-003 executes which loads the debug code into cog $1F0-1F3.
'   The debug code executes next, which reloads the original (saved) user code back into cog $000-003.
'   The debug code sets the first instruction for the cog to be $000 and waits for the dubugger code
'     to tell it what to do.
'   Execution is controlled by a simple LMM kernel.
'------------------------------------------------------------------------------------------------------------
Debug_Block             org     0                       '  v0.275   
SAVE_000                nop                             '\ Cog user code $000-003 is saved here
SAVE_001                nop                             '|
SAVE_002                nop                             '|
SAVE_003                nop                             '/
                        org     0
' Debug LMM code executes from Hub (works cooperatively with the Debug LMM code in Cog)
D_LMM_BOOT              rdlong  X_ENTER, #DB+@z_ENTER   '\ copies Debug code
                        rdlong  X_ADD4,  #DB+@z_ADD4    '|   from hub to cog $1F0-1F3
                        rdlong  X_OPCODE,#DB+@z_OPCODE  '| (workaround compiler restriction)
                        rdlong  X_1F3,   #DB+@z_1F3     '/
                        jmp     #X_ENTER
                        org     0
' This is the LMM debug code for the Debug kernel running in cog $1F0-1F3
D_LMM_SAVE              rdlong  SAVE_000,#DB+@SAVE_000  '\ restores the user code
                        rdlong  SAVE_001,#DB+@SAVE_001  '|   from hub to cog $000-003
                        rdlong  SAVE_002,#DB+@SAVE_002  '|
                        rdlong  SAVE_003,#DB+@SAVE_003  '/
' This is the LMM debug execution loop which the Debug kernel will then execute & communicate with the spin debugger
D_LMM_EXEC              sub     X_ENTER,#4-0            'normally #4 (set to #0 by "execute")
D_OPC_EXEC              nop                             'placeholder for the instruction to execute
                        wrbyte  X_ADD4,#DB+@D_LMM_EXEC  'set back to #4 when done
                        movs    X_ENTER,#DB+@D_LMM_EXEC 'simulates jump to D_LMM by changing the hub pointer
' The following are used to store variable data for hub store/load
D_VAL1_EXEC             long    0                       'data value  for cog to hub store/load
D_VAL2_EXEC             long    0                       'data value2 for cog to hub store/load (not used yet)
D_VAL3_EXEC             long    0                       'data value3 for cog to hub store/load (not used yet)
' Debug kernel code runs in cog shadow ram $1F0-1F3
                        org     $0                      'compiler will not allow $1F0 so workaround
z_ENTER                 rdlong  X_OPCODE,#DB+@D_LMM_SAVE
z_ADD4                  add     X_ENTER,#4
z_OPCODE                nop
z_1F3                   jmp     #X_ENTER
'------------------------------------------------------------------------------------------------------------
' Debug kernel code runs in cog $000-003 (bootloader code executes when cog starts)
'   The user code in cog $000-003 is replaced with the code below by spin while still in hub memory.
'   The user code is restored once the Debug kernel is loaded into cog $1F0-1F3.
                        org     0
B_ENTER                 rdlong  B_OPCODE,#DB+@D_LMM_BOOT
B_ADD4                  add     B_ENTER,#4
B_OPCODE                nop
B_003                   jmp     #B_ENTER
'<><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>

OBJ
	dbg	: "ClusoDebugger_276"
}
VAR
	long	cog

PUB start(params, cpu_regs) : okay
	stop
	iocmd_ptr := long[params][0]
#ifdef BANKED_MEM
	bankbase_ptr := long[params][0] + 4
#endif
#ifdef INTERRUPTS
	irq_ptr := long[params][1]
#endif
#ifdef MICROTAN
	screen_ptr := long[params][2]
	graphics_ptr := long[params][3]
#endif MICROTAN
{
	dbg.StartDebugger(-1, @Debug_Block, @entry, 0)
	waitcnt(clkfreq / 1000 * 200 + cnt)
}
	COGNEW(@entry, cpu_regs)

PUB stop
	if cog
		COGSTOP(cog~ - 1)

DAT
			org	0
entry

#ifdef TriBladeProp
'The TriBladeProp Blade #2 driver presets the latch to enable SRAM -CE(U23)=0 and hold LE=0
outx			mov     outa, ram_latched		' -WE=1 -OE=1, all other bits=0
datax			mov     dira, ram_dir_read		' setup direction to read
			jmp	#startup
#endif TriBladeProp

#ifdef RamBladeProp
'The RamBladeProp needs this?
outx			mov     outa, #0
datax			jmp	#startup
#endif RamBladeProp

#ifdef DracBladeProp
outx			jmp	#startup
#endif

'***************************************************************************************************************
'
' Write back alu to effective address
' (NB: wb_ea must be an address with only 5 bits, i.e. $00..$1f)
'
wb_ea
			call	#wr_byte
			jmp	#jmp_vector


'***************************************************************************************************************
'
' Write back alu to accumulator
' (NB: wb_acc must be an address with only 5 bits, i.e. $00..$1f)
'
wb_acc
			mov	A, alu
			jmp	#jmp_vector


startup
			mov	t1, PAR
			rdword	PC, t1
			add	t1, #2
			rdbyte	PSW, t1
			add	t1, #1
			rdbyte	A, t1
			add	t1, #1
			rdbyte	X, t1
			add	t1, #1
			rdbyte	Y, t1
			add	t1, #1
			rdbyte	S, t1

'***************************************************************************************************************
' No OPeration
'
m_nop

'***************************************************************************************************************
' Jump to next address in vector
'
jmp_vector
			shr	vector, #9		wz
		if_nz	jmp	vector

'***************************************************************************************************************
' Fetch opcode and dispatch through op_table
'
fetch
#ifdef INTERRUPTS
			rdlong	t1, irq_ptr		wz
		if_nz	mov	lmm_pc, do_int
		if_nz	call	#lmm_next
#endif INTERRUPTS
fetch_0
			call	#rd_opcode
			mov	opcode, alu
			shl	alu, #2
			add	alu, op_table
			rdlong	vector, alu
			jmp	vector

#ifdef PAGED_MEMORY
mem_page_256
			and	ea, low_word
			mov	outx, ea
			shr	outx, #8			' discard 8 bits of the address
mem_table		add	outx, 0-0			' source is patched with the table address
			rdbyte	outx, outx			' fetch memory access handler
			jmp	outx				' and jump there
rd_nop
			mov	alu, #$ff			' read nothing but FF
			jmp	rd_byte_ret
rd_i_o
			mov	outx, #io_cmd_in
			shl	outx, #16
			or	outx, ea
			shl	outx, #8
			wrlong	outx, iocmd_ptr
:wait			rdlong	outx, iocmd_ptr
			shr	outx, #24		wz, nr
		if_nz	jmp	#:wait
			mov	alu, outx
			and	alu, #$ff
			jmp	rd_byte_ret
wr_i_o
			mov	outx, #io_cmd_out
			shl	outx, #16
			or	outx, ea
			shl	outx, #8
			or	outx, alu
			wrlong	outx, iocmd_ptr
:wait			rdlong	outx, iocmd_ptr
			shr	outx, #24		wz, nr
		if_nz	jmp	#:wait
			jmp	#wr_ram
#endif PAGED_MEMORY

'***************************************************************************************************************
'
'	Memory access functions (TriBladeProp)
'
'***************************************************************************************************************
#ifdef TriBladeProp
'                                          +-------- -CS (U26 FLASH)   -+                                  
'                                          |+------- -CS (J22 microSD)  +- Latched pins...                 
'                                          ||+------ -RST3              |    passes when LE=1              
'             (P30) SO ---+                |||+----- -RST1              |    latches on  LE low going edge 
'             (P31) SI --+|                ||||+---- -CE (U24 SRAM-hi)  |    latched on  LE=0              
'                        ||+---- -WE(SDA)  |||||+--- -CE (U23 SRAM-lo)  |                                  
'                        |||+--- -OE(SCL)  ||||||+-- A20                |                                  
'                        ||||+--  LE       |||||||+- A19               -+                                  
'                        |||||             ||||||||                                                        
'          P31...P0 --> %00xx0_xxxxxxxxxxx_xxxxxxxx_xxxxxxxx  
ram_latched	long	%00110_00000000000_00000000_00000000	' -WE=1 -OE=1 (LE=0)                               
ram_read	long	%00100_00000000000_00000000_00000000	' -WE=1 -OE=0 (LE=0)    
ram_write	long	%00010_00000000000_00000000_00000000	' -WE=0 -OE=1 (LE=0)
ram_dir_read	long	%00110_11111111111_11111111_00000000	' outputs WE, OE, A0-18, inputs D0-7
ram_dir_write	long	%00110_11111111111_11111111_11111111	' outputs WE, OE, A0-18, D0-7
ram_dir_input	long	%00000_00000000000_00000000_00000000	' all inputs (for Ram Disk access)
'                              +---------+ +------+ +------+
'                              A18......A8 A7....A0 D7....D0
'                                          Q7....Q0
'
'***************************************************************************************************************
' Read a byte from RAM[ea] into the alu
'
rd_byte
#ifdef BANKED_MEM
			cmp	ea, banklimit		wc
		if_c	add	ea, bankbase
#endif BANKED_MEM
#ifdef PAGED_MEMORY
			movs	mem_table, #rd_memory_ptr
			jmp	#mem_page_256
rd_rom
rd_ram
#endif PAGED_MEMORY
			mov	outx, ea
			shl	outx, #8			' shift address 8 bits
			or	outx, ram_read			' add -WE=1 -OE=0
			mov	outa, outx			' pc (shifted 8 bits + -WE=1 -OE=0)
			nop					' delay for sram access time
			mov     alu, ina			' read SRAM
			and     alu, #$ff			' extract 8 bits
rd_byte_ret
			ret

'***************************************************************************************************************
' Write a byte from the alu into RAM[ea]
'
wr_byte
#ifdef BANKED_MEM
			cmp	ea, banklimit		wc
		if_c	add	ea, bankbase
#endif BANKED_MEM
#ifdef PAGED_MEMORY
			movs	mem_table, #wr_memory_ptr	' dispatch according to the wr_memory table
			jmp	#mem_page_256			' in 256 byte pages
wr_ram
#endif PAGED_MEMORY
			mov	outx, ea
			shl	outx, #8			' shift address 8 bits
			or	outx, ram_write			' add -WE=0 -OE=1
			mov	datax, alu
			and	datax, #$ff			' ensure upper bytes=0
			or	outx, datax			' add data bits to be written
			mov	outa, outx			' output -WE=0 -OE=1, address bits, data bits
			mov	dira, ram_dir_write		' set data bits to outputs
			or      outa, ram_latched		' set -WE=0->1 -OE=1->1  (end write)
			mov     dira, ram_dir_read		' set data bits to inputs
			mov     outa, ram_read			' set -WE=1 -OE=0
#ifdef PAGED_MEMORY
wr_rom
wr_nop
#endif PAGED_MEMORY
wr_byte_ret
			ret
#endif TriBladeProp





'***************************************************************************************************************
'
'	Memory access functions (RamBladeProp)
'
'***************************************************************************************************************
#ifdef RamBladeProp
' RamBlade Hardware Access:
'       P0-18  = A0-18   SRAM address
'       P19    = -CS     microSD
'       P20    = -CE     SRAM & enable eeprom
'       P21    = -WE     SRAM & enable eeprom
'       P22    = SO      \ UART
'       P23    = SI      /
'       P24-31 = D0-7    SRAM \ * see below
'                             | P28 = SCL eeprom
'                             | P29 = SDA eeprom
'                             | P30 = prog SO
'                             / P31 = prog SI
'
'         (microSD) DO ---------+
'         (microSD) DI --------+| +--------  SI (P23)
'         (microSD) CLK-------+|| |+-------  SO (P22)
'                             ||| ||
'             (P28) SCL ----+ ||| ||+------ -WE (SRAM & enable eeprom)
'             (P29) SDA ---+| ||| |||+----- -CE (SRAM & enable eeprom)
'             (P30) SO ---+|| ||| ||||+---- -CS (microSD)
'             (P31) SI --+||| ||| |||||
'                        |||| ||| |||||
'          P31...P0 --> %xxxxxxxx_00xxx_xxxxxxxxxxxxxxxxxxx
'       10K pullups -->    ** ***   ***
'                        +------+       +-----------------+
'                        D7....D0       A18..............A0
ram_dir_read    long    %00000000_00010_1111111111111111111 ' outputs CE, A0-18           (sram read)
ram_dir_write   long    %11111111_00000_1111111111111111111 ' outputs A0-18, D0-7         (sram disabled)
ram_dir_wstrobe long    %11111111_00110_1111111111111111111 ' outputs CE, WE, A0-19, D0-7 (write strobe)
'
'***************************************************************************************************************
' Read a byte from RAM[ea] into the alu
'
rd_byte
#ifdef BANKED_MEM
			cmp	ea, banklimit		wc
		if_c	add	ea, bankbase
#endif BANKED_MEM
#ifdef PAGED_MEMORY
			movs	mem_table, #rd_memory_ptr
			jmp	#mem_page_256
rd_rom
rd_ram
#endif PAGED_MEMORY
			mov	OUTA, ea			' apply the read address incl -CE=0 -WE=0
			mov	DIRA, ram_dir_read		' set outputs A0-18, CE
			nop					' sram delay required
			mov	alu, INA			' read byte from SRAM
			shr	alu, #24			' bits 31..24 to 7..0
			mov	DIRA, #0			' tri state
rd_byte_ret
			ret

'***************************************************************************************************************
' Write a byte from the alu into RAM[ea]
'
wr_byte
#ifdef BANKED_MEM
			cmp	ea, banklimit		wc
		if_c	add	ea, bankbase
#endif BANKED_MEM
#ifdef PAGED_MEMORY
			movs	mem_table, #wr_memory_ptr
			jmp	#mem_page_256
wr_ram
#endif PAGED_MEMORY
			shl	alu, #24			' data byte in bits 31..24
			or	alu, ea				' data in 31..24, address in 18..0
			or	OUTA, alu
			mov	DIRA, ram_dir_write		' set outputs A0-18
			mov	DIRA, ram_dir_wstrobe		' -CE=-WE= 1 -> 0       \ ce + we pulse
			shr	alu, #24			' data byte in bits 7..0 | delay
			mov	DIRA, ram_dir_write		' address + data only   /
			mov	DIRA, #0			' tri state
#ifdef PAGED_MEMORY
wr_rom
wr_nop
#endif PAGED_MEMORY
wr_byte_ret
			ret
#endif RamBladeProp





'***************************************************************************************************************
'
'	Memory access functions (DracBladeProp)
'
'***************************************************************************************************************
#ifdef DracBladeProp
'****************** Dr_Acula single propeller driver **************************
' Three latches to access the ram chip - see main program for spin version
' uses variables 'ea' and 'alu'. skip banked memory and checking for out of
' bounds address will need two nop instructions in /rd and /wr to be sure 
'
dir_138_read	long    %00000000_00000000_00001111_00000000	' for reads so data lines are tristate till the read
dir_138_write	long    %00000000_00000000_00001111_11111111	' for writes so data lines are outputs
address_low	long    %00000000_00000000_00000100_00000000	' low address latch = xxxx010x and gate low xxxxxxx0
address_mid	long    %00000000_00000000_00000110_00000000	' middle address latch = xxxx011x and gate low xxxxxxx0
address_high	long    %00000000_00000000_00001000_00000000	' highe address latch = xxxx100x and gate low xxxxxxx0
write_enable	long    %00000000_00000000_00000010_00000000	' write enable xxxx001x and gate low xxxxxxx0
'
'***************************************************************************************************************
' Read a byte from RAM[ea] into the alu
'
rd_byte
#ifdef PAGED_MEMORY
			movs	mem_table, #rd_memory_ptr
			jmp	#mem_page_256
rd_rom
rd_ram
#endif PAGED_MEMORY
			call	#ram_address			' sets up the latches with the correct ram address
			mov	DIRA, dir_138_read		' for reads so P0-P7 tristate till do read
			mov	OUTA, #$0			' set 000 low 138 read, low gate and P0-P7 are inputs
			nop					' none does not work, 1 does work
			mov	alu, INA			' read SRAM
			and	alu, #$ff			' extract 8 bits
			or	OUTA, #%1_00000000		' set the gate high again
			mov	DIRA, #0			' tristate all pins
rd_byte_ret
			ret

'***************************************************************************************************************
' Write a byte from the alu into RAM[ea]
'
wr_byte
#ifdef PAGED_MEMORY
			movs	mem_table, #wr_memory_ptr
			jmp	#mem_page_256
wr_ram
#endif PAGED_MEMORY
        		call	#ram_address			' sets up the latches with the correct ram address
        		mov	outx, alu			' get the byte to output
        		and	outx, #$ff			' ensure upper bytes=0
        		or	outx, write_enable		' or with correct 138 address
	        	mov	OUTA, outx			' send it out
	        	nop					' NOP  needs at least one due to 138 propogation delay
	        	or	OUTA, #%1_00000000		' set it high again
	        	mov	DIRA, #0			' tristate all pins
#ifdef PAGED_MEMORY
wr_rom
wr_nop
#endif PAGED_MEMORY
wr_byte_ret
			ret

'***************************************************************************************************************
' Sets up the ram latches.
' Assumes high latch A16-A18 low so only
' accesses 64K of RAM
ram_address
			and	ea, low_word
#ifdef BANKED_MEM
			cmp	ea, banklimit		wc
		if_c	add	ea, bankbase
#endif BANKED_MEM
			mov	DIRA, dir_138_write		' set up the pins for programming latch chips
			mov	outx, ea			' get the address into a temp variable
			and	outx, #$ff			' mask the low byte
			or	outx, address_low		' or with 138 low address
			mov	OUTA, outx			' send it out with gate low
			mov	outx, ea			' get the address into a temp variable
			or	OUTA, #%1_00000000		' set gate high again and repeat for the middle byte
			shr	outx, #8			' shift right by 8 places
			and	outx, #$ff			' mask the low byte
			or	outx, address_mid		' or with 138 middle address
			mov	OUTA, outx			' send it out wit gate low
			mov	outx, ea			' get the address into a temp variable
			or	OUTA, #%1_00000000		' set gate high again
#ifdef BANKED_MEM
			shr	outx, #16			' shift right by 16 places
			and	outx, #$07			' mask 3 bits in the high byte
			or	outx, address_high		' or with 138 high address
			mov	OUTA, outx			' send it out with gate low
			or	OUTA, #%1_00000000		' set gate high again
#endif BANKED_MEM
ram_address_ret
			ret
#endif DracBladeProp




'******************************************************************************
'
' Pop a byte off the stack
'
pull_alu
			add	S, #1
			and	S, #$ff
			mov	ea, S
			or	ea, #$100
			call	#rd_byte
pull_alu_ret
			ret

'******************************************************************************
'
' Push a byte to the stack
'
push_alu
			mov	ea, S
			or	ea, #$100
			call	#wr_byte
			sub	S, #1
			and	S, #$ff
push_alu_ret
			ret

'***************************************************************************************************************
'
' Push program counter
'
push_pc
			call	#rd_stack			' for emulation accuracy's sake: read memory (SP)
			mov	alu, PC
			shr	alu, #8
			call	#push_alu
			mov	alu, PC
			and	alu, #$ff
			call	#push_alu
push_pc_ret
			ret

'***************************************************************************************************************
'
' Read a byte from address [PC++] into alu
'
rd_opcode
			mov	ea, PC
			add	PC, #1
			and	PC, low_word
			call	#rd_byte
rd_opcode_ret
			ret

'***************************************************************************************************************
'
' Read a byte from address [SP] into alu
'
rd_stack
			mov	ea, S
			or	ea, #$100
			call	#rd_byte
rd_stack_ret
			ret

'***************************************************************************************************************
'
' Read a word from address [PC++] into alu
'
rd_opword
			mov	ea, PC
			call	#rd_byte
			movs	:or, alu
			add	ea, #1
			and	ea, low_word
			call	#rd_byte
			add	ea, #1
			and	ea, low_word
			mov	PC, ea
			shl	alu, #8
:or			or	alu, #0-0
rd_opword_ret
			ret

'***************************************************************************************************************
'
' Calculate an absolute address plus X/Y; additional read if a page boundary is crossed
'
calc_abs_xy_p
			call	#rd_opword		' get absolute address
			mov	ea, alu
			and	alu, #$ff
			add	alu, XY
			test	alu, #$100	wc	' page wrap?
			add	ea, XY
			and	ea, low_word		' stay inside 64K
		if_nc	jmp	calc_abs_xy_p_ret	' no, just return (no #, we jump to the source field of the ret)
			sub	ea, #$100		' original page
			and	ea, low_word
			call	#rd_byte		' read wrong page byte
			add	ea, #$100		' correct page
			and	ea, low_word
calc_abs_xy_p_ret
			ret

'***************************************************************************************************************
'
' Calculate an absolute address plus X; stays inside page
'
calc_abs_xy_np
			call	#rd_opword
			mov	ea, alu
			andn	ea, #$ff
			add	alu, XY
			and	alu, #$ff
			or	ea, alu
calc_abs_xy_np_ret
			ret

'***************************************************************************************************************
'
' EA = zero page + X indirect (pre indexed)
'
calc_idx
			call	#rd_opcode		' zero page address
i_add_alu_x		add	alu, X
			and	alu, #$ff
			mov	ea, alu
			call	#rd_byte		' read LSB
			movs	:or, alu		' store LSB to OR to address
			add	ea, #1
			and	ea, #$ff
			call	#rd_byte		' read MSB
			shl	alu, #8
:or			or	alu, #0-0
			mov	ea, alu
calc_idx_ret
			ret

'***************************************************************************************************************
'
' Calculate zero page indirect + Y (post indexed); stays inside page
'
calc_idy_np
			call	#rd_opcode		' zero page address
			mov	ea, alu
			call	#rd_byte		' read LSB
			add	alu, Y
			mov	t1, alu
			add	ea, #1
			and	ea, #$ff
			call	#rd_byte		' read MSB
			shl	alu, #8
			add	alu, t1
			and	t1, #$100	wz	' t1 = $100 if the page wrapped
			sub	alu, t1			' stay in original page
			mov	ea, alu			' effective address
calc_idy_np_ret
			ret

'***************************************************************************************************************
'
' Calculate zero page indirect + Y (post indexed); additional read if a page boundary is crossed
'
calc_idy_p
			call	#calc_idy_np
		if_z	jmp	calc_idy_p_ret		' no page wrap (no #, we jump to the source field of the ret)
			call	#rd_byte		' read wrong page byte
			add	ea, #$100
			and	ea, low_word		' stay inside 64K
calc_idy_p_ret
			ret

'***************************************************************************************************************
'
' Read accumulator
'
rd_acc
			mov	alu, A
			jmp	#jmp_vector

'***************************************************************************************************************
'
' Read an immediate value
'
rd_imm
			call	#rd_opcode
			jmp	#jmp_vector

'***************************************************************************************************************
'
' Dummy read the next PC byte
'
rd_dummy
			mov	ea, PC
			call	#rd_byte
			jmp	#jmp_vector

'***************************************************************************************************************
'
' Read a zero page address
'
rd_zpg
			mov	rd_zpg_add, i_nop
rd_zpg_0
			call	#rd_opcode
rd_zpg_add		nop
			and	alu, #$ff
			mov	ea, alu
rd_alu
			movs	rd_byte_ret, #jmp_vector
			jmp	#rd_byte

'***************************************************************************************************************
'
' Write a zero page address
'
wr_zpg
			mov	wr_zpg_add, i_nop
wr_zpg_0
			call	#rd_opcode
wr_zpg_add		nop
			and	alu, #$ff
			mov	ea, alu
wr_mem
			mov	alu, mem
			movs	wr_byte_ret, #fetch
			jmp	#wr_byte

'***************************************************************************************************************
'
' Read a zero page address plus X
'
rd_zpx
			mov	rd_zpg_add, i_add_alu_x
			jmp	#rd_zpg_0

'***************************************************************************************************************
'
' Write a zero page address plus X
'
wr_zpx
			mov	wr_zpg_add, i_add_alu_x
			jmp	#wr_zpg_0

'***************************************************************************************************************
'
' Read a zero page address plus y
'
rd_zpy
			mov	rd_zpg_add, i_add_alu_y
			jmp	#rd_zpg_0

'***************************************************************************************************************
'
' Write a zero page address plus Y
'
wr_zpy
			mov	wr_zpg_add, i_add_alu_y
			jmp	#wr_zpg_0

'***************************************************************************************************************
'
' EA = an absolute address
'
ea_abs
			call	#rd_opword
			mov	ea, alu
			jmp	#jmp_vector

'***************************************************************************************************************
'
' Read an absolute address
'
rd_abs
			call	#rd_opword
			mov	ea, alu
			jmp	#rd_alu

'***************************************************************************************************************
'
' Write an absolute address
'
wr_abs
			call	#rd_opword
			mov	ea, alu
			jmp	#wr_mem

'***************************************************************************************************************
'
' Read byte from an absolute address plus X; additional read if a page boundary is crossed
'
rd_abx_p
			mov	XY, X
			call	#calc_abs_xy_p
			jmp	#rd_alu

'***************************************************************************************************************
'
' Read byte from an absolute address plus X; stays inside page
'
rd_abx_np
			mov	XY, X
			movs	calc_abs_xy_np_ret, #rd_alu
			jmp	#calc_abs_xy_np

'***************************************************************************************************************
'
' Write byte to an absolute address plus X; stays inside page
'
wr_abx_np
			mov	XY, X
			movs	calc_abs_xy_np_ret, #wr_mem
			jmp	#calc_abs_xy_np

'***************************************************************************************************************
'
' Read byte from an absolute address plus Y; additional read if a page boundary is crossed
'
rd_aby_p
			mov	XY, Y
			movs	calc_abs_xy_p_ret, #rd_alu
			jmp	#calc_abs_xy_p

'***************************************************************************************************************
'
' Read byte from an absolute address plus Y; stays inside page
'
rd_aby_np
			mov	XY, Y
			movs	calc_abs_xy_np_ret, #rd_alu
			jmp	#calc_abs_xy_np

'***************************************************************************************************************
'
' Write byte to an absolute address plus Y; stays inside page
'
wr_aby_np
			mov	XY, Y
			movs	calc_abs_xy_np_ret, #wr_mem
			jmp	#calc_abs_xy_np

'***************************************************************************************************************
'
' Read byte from zero page + X indirect (pre indexed)
'
rd_idx
			movs	calc_idx_ret, #rd_alu
			jmp	#calc_idx

'***************************************************************************************************************
'
' Write byte to zero page + X indirect (pre indexed)
'
wr_idx
			movs	calc_idx_ret, #wr_mem
			jmp	#calc_idx

'***************************************************************************************************************
'
' Read byte from address zero page indirect + Y (post indexed); additional read if a page boundary is crossed
'
rd_idy_p
			movs	calc_idy_p_ret, #rd_alu
			jmp	#calc_idy_p

'***************************************************************************************************************
'
' Read byte from address zero page indirect + Y (post indexed); stays in same page
'
rd_idy_np
			movs	calc_idy_np_ret, #rd_alu
			jmp	#calc_idy_np

'***************************************************************************************************************
'
' Write byte to address zero page indirect + Y (post indexed); stays in same page
'
wr_idy_np
			movs	calc_idy_np_ret, #wr_mem
			jmp	#calc_idy_np

'***************************************************************************************************************
'
' Calculate negative and zero flags of alu
'
nz_flags
			test	alu, #$80		wc
			and	alu, #$ff		wz
			muxc	PSW, #NF
			muxz	PSW, #ZF
nz_flags_ret
			ret

'***************************************************************************************************************
'
' BMI/BCS/BEQ/BVS: Branch if flag is set
'
bra_set
			call	#rd_opcode
			shr	vector, #9
			test	vector, PSW		wz
		if_z	jmp	#fetch
			jmp	#m_bra

'***************************************************************************************************************
'
' BPL/BCC/BNE/BVC: Branch if flag is clear
'
bra_clr
			call	#rd_opcode
			shr	vector, #9
			testn	vector, PSW		wz
		if_z	jmp	#fetch
			' fall through

'***************************************************************************************************************
'
' Branch relative to PC
'
m_bra
			shl	alu, #24			' sign extend alu
			sar	alu, #24
			add	PC, alu				' add to PC
			and	PC, low_word
			jmp	#fetch

'***************************************************************************************************************
'
' PHP: Push processor status
'
m_php
			mov	alu, PSW
			movs	push_alu_ret, #fetch
			jmp	#push_alu

'***************************************************************************************************************
'
' PLP: Pull processor status
'
m_plp
			call	#rd_stack			' for emulation accuracy's sake: read memory (SP)
			call	#pull_alu
			or	alu, #(TF | BF)			' set break and trace flags (???)
#ifdef INTERRUPTS
			test	PSW, #II		wz
			mov	PSW, alu
			test	PSW, #II		wc	
	if_nz_and_nc	jmp	#fetch_0			' no interrupt immediately after PLP that clears II
			jmp	#fetch
#else  INTERRUPTS
			mov	PSW, alu
			jmp	#fetch
#endif !INTERRUPTS

'***************************************************************************************************************
'
' PHA: Push accumulator
'
m_pha
			mov	alu, A
			movs	push_alu_ret, #fetch
			jmp	#push_alu

'***************************************************************************************************************
'
' PLA: Pull accumulator
'
m_pla
			call	#rd_stack			' for emulation accuracy's sake: read memory (SP)
			call	#pull_alu
			mov	A, alu
			movs	nz_flags_ret, #fetch
			jmp	#nz_flags

'***************************************************************************************************************
'
' TXA: Transfer X to accumulator
'
m_txa
			mov	A, X
			mov	alu, A
			movs	nz_flags_ret, #fetch
			jmp	#nz_flags

'***************************************************************************************************************
'
' TAX: Transfer accumulator to X
'
m_tax
			mov	X, A
			mov	alu, X
			movs	nz_flags_ret, #fetch
			jmp	#nz_flags

'***************************************************************************************************************
'
' TXS: Transfer X to stack pointer (no flags effect!)
'
m_txs
			mov	S, X
			jmp	#fetch

'***************************************************************************************************************
'
' TSX: Transfer stack pointer to X
'
m_tsx
			mov	X, S
			mov	alu, X
			movs	nz_flags_ret, #fetch
			jmp	#nz_flags

'***************************************************************************************************************
'
' TYA: Transfer Y to accumulator
'
m_tya
			mov	A, Y
			mov	alu, A
			movs	nz_flags_ret, #fetch
			jmp	#nz_flags

'***************************************************************************************************************
'
' TAY: Transfer accumulator to Y
'
m_tay
			mov	Y, A
			mov	alu, Y
			movs	nz_flags_ret, #fetch
			jmp	#nz_flags

'***************************************************************************************************************
'
' CLC/CLD/CLI/CLV: Clear flag (vector >> 9)
'
clr_flag
			shr	vector, #9
			andn	PSW, vector
			jmp	#fetch

'***************************************************************************************************************
'
' SEC/SED/SEI: Set flag (vector >> 9)
'
set_flag
			shr	vector, #9
			or	PSW, vector
			jmp	#fetch

'***************************************************************************************************************
'
' LDA: Load accumulator
'
m_lda
			mov	A, alu
			movs	nz_flags_ret, #fetch
			jmp	#nz_flags

'***************************************************************************************************************
'
' (6510) LAX: Load accumulator and X register
'
m_lax
			mov	A, alu
			' fall through

'***************************************************************************************************************
'
' LDX: Load X register
'
m_ldx
			mov	X, alu
			movs	nz_flags_ret, #fetch
			jmp	#nz_flags

'***************************************************************************************************************
'
' LDY: Load Y register
'
m_ldy
			mov	Y, alu
			movs	nz_flags_ret, #fetch
			jmp	#nz_flags

'***************************************************************************************************************
'
' STA: Store accumulator
'
m_sta
			mov	mem, A
			jmp	#jmp_vector

'***************************************************************************************************************
'
' STX: Store X register
'
m_stx
			mov	mem, X
			jmp	#jmp_vector

'***************************************************************************************************************
'
' (6510) SAX: Store A & X register
'
m_sax
			mov	mem, A
			and	mem, X
			jmp	#jmp_vector

'***************************************************************************************************************
'
' STY: Store Y register
'
m_sty
			mov	mem, Y
			jmp	#jmp_vector

'***************************************************************************************************************
'
' AND: Logical AND accumulator
'
alu_and
			and	alu, A
			movs	nz_flags_ret, #wb_acc
			jmp	#nz_flags

'***************************************************************************************************************
'
' (6510) ANC: Logical AND accumulator, set carry from bit 7
'
alu_anc
			and	alu, A
			test	alu, #%10000000		wc
			muxc	PSW, #CF
			movs	nz_flags_ret, #wb_acc
			jmp	#nz_flags

'***************************************************************************************************************
'
' ORA: Logical OR accumulator
' NB: resetting the bits of A in alu turns the following EOR into an ORA
'
alu_ora
			andn	alu, A
			' fall through

'***************************************************************************************************************
'
' EOR: Logical Exclusive OR accumulator
'
alu_eor
			xor	alu, A
			movs	nz_flags_ret, #wb_acc
			jmp	#nz_flags
		
'***************************************************************************************************************
'
' (6510) SLO: Arithmetic shift left and OR
' C <- [7][6][5][4][3][2][1][0] <- 0
'
alu_slo
			shl	alu, #1
			test	alu, #$100		wc
			muxc	PSW, #CF
			or	alu, A
			and	alu, #$ff
			mov	A, alu
			mov	mem, alu
			movs	nz_flags_ret, #jmp_vector
			jmp	#nz_flags

'***************************************************************************************************************
'
' ASL: Arithmetic shift left
' C <- [7][6][5][4][3][2][1][0] <- 0
'
alu_asl
			shl	alu, #1
			test	alu, #$100		wc
			muxc	PSW, #CF
alu_tmp_ff
			and	alu, #$ff
			mov	mem, alu
			movs	nz_flags_ret, #jmp_vector
			jmp	#nz_flags

'***************************************************************************************************************
'
' (6510) ASR: And with immediate and logic shift right
'
alu_asr
			and	alu, A

'***************************************************************************************************************
'
' LSR: Logic shift right
' 0 -> [7][6][5][4][3][2][1][0] -> C
'
alu_lsr
			shr	alu, #1			wc
			muxc	PSW, #CF
			jmp	#alu_tmp_ff

'***************************************************************************************************************
'
' (6510) SRE: Logic shift right and exclusive OR
' 0 -> [7][6][5][4][3][2][1][0] -> C
'
alu_sre
			shr	alu, #1			wc
			muxc	PSW, #CF
			xor	alu, A
			jmp	#alu_tmp_ff

'***************************************************************************************************************
'
' ROL: Rotate left
' new C <- [7][6][5][4][3][2][1][0] <- C
'
alu_rol
			test	PSW, #CF		wc
			rcl	alu, #1
			test	alu, #$100		wc
			muxc	PSW, #CF
			jmp	#alu_tmp_ff

'***************************************************************************************************************
'
' ROR: Rotate right
' C -> [7][6][5][4][3][2][1][0] -> new C
'
alu_ror
			test	PSW, #CF		wz
			shr	alu, #1			wc
			muxc	PSW, #CF
			muxnz	alu, #%10000000
			jmp	#alu_tmp_ff

'***************************************************************************************************************
'
' BIT: Bit test accumulator
'
alu_bit
			andn	PSW, #(NF | VF)			' clear negative and overflow flags
			mov	aux, alu
			and	aux, #(NF | VF)			' mask negative and overflow flags
			or	PSW, aux			' set in PSW
			test	A, mem			wz	' test bit(s)
			muxz	PSW, #ZF			' set zero flag if none set
			jmp	#fetch

'***************************************************************************************************************
'
' JMP: Jump to effective address
'
m_jmp
			mov	PC, ea
			jmp	#fetch

'***************************************************************************************************************
'
' INC: Increment memory
'
alu_inc
			add	alu, #1
			jmp	#alu_tmp_ff

'***************************************************************************************************************
'
' DEC: Decrement memory
'
alu_dec
			sub	alu, #1
			jmp	#alu_tmp_ff

'***************************************************************************************************************
'
' INX: Increment X register
'
m_inx
			add	X, #1
			mov	alu, X
			movs	nz_flags_ret, #fetch
			jmp	#nz_flags

'***************************************************************************************************************
'
' INY: Increment Y register
'
m_iny
			add	Y, #1
			mov	alu, Y
			movs	nz_flags_ret, #fetch
			jmp	#nz_flags

'***************************************************************************************************************
'
' DEX: Decrement X register
'
m_dex
			sub	X, #1
			mov	alu, X
			movs	nz_flags_ret, #fetch
			jmp	#nz_flags

'***************************************************************************************************************
'
' DEY: Decrement Y register
'
m_dey
			sub	Y, #1
			mov	alu, Y
			movs	nz_flags_ret, #fetch
			jmp	#nz_flags

'***************************************************************************************************************
'
' (6510) DCP: Decrement and compare accumulator
'
alu_dcp
			sub	alu, #1
			and	alu, #$ff
			' fall through

'***************************************************************************************************************
'
' CMP: Compare accumulator with alu
'
alu_cmp
			mov	mem, alu
			mov	alu, A
alu_cmp_1		sub	alu, mem		wc
			muxnc	PSW, #CF
			movs	nz_flags_ret, #fetch
			jmp	#nz_flags

'***************************************************************************************************************
'
' CPX: Compare X register
'
m_cpx
			mov	mem, alu
			mov	alu, X
			jmp	#alu_cmp_1

'***************************************************************************************************************
'
' CPY: Compare X register
'
m_cpy
			mov	mem, alu
			mov	alu, Y
			jmp	#alu_cmp_1

'***************************************************************************************************************
'
' Execute an LMM function
'
lmm
			shr	vector, #9
lmm_1
			add	vector, lmm_table
			rdlong	lmm_pc, vector
lmm_next		rdlong	:op1, lmm_pc
			add	lmm_pc, #4
:op1			nop
			rdlong	:op2, lmm_pc
			add	lmm_pc, #4
:op2			nop
			rdlong	:op3, lmm_pc
			add	lmm_pc, #4
:op3			nop
			rdlong	:op4, lmm_pc
			add	lmm_pc, #4
:op4			nop
			jmp	#lmm_next
lmm_next_ret
lmm_1_ret
lmm_ret
			ret

#ifdef INTERRUPTS
irq_ptr			long	0-0				' interrupt request address
do_int			long	@@@do_interrupt
#endif INTERRUPTS

op_table		long	@@@opcodes
lmm_table		long	@@@lmm_functions

#ifdef PAGED_MEMORY
rd_memory_ptr		long	@@@rd_memory
wr_memory_ptr		long	@@@wr_memory
screen_ptr		long	0-0
graphics_ptr		long	0-0
#endif

#ifdef BANKED_MEM
bankbase_ptr		long	0-0
banklimit		long	$c000
bankbase		long	0
#endif BANKED_MEM

iocmd_ptr		long	0-0				' set by Spin code

context_ptr		long	@@@PC_reg			' PC register address
trace_ptr		long	@@@lmm_trace
low_word		long	$0000ffff
cflag			long	CF				' need C flag as destination operand

i_nop			nop
i_add_alu_y		add	alu, Y

lo			long	0				' temporary lo nibble
hi			long	0				' temporary hi nibble
t1			long	0				' temporary address storage

PC			long	0				' program counter
PSW			long	II | ZF				' processor status (reset value is zero flag set)
A			long	0				' accumulator
X			long	0				' X index register
Y			long	0				' Y index register
S			long	0				' stack pointer (LSB)
XY			long	0				' one of X or Y

ea			res	1				' effective address
opcode			res	1				' most recent opcode
alu			res	1				' arithmetic logic unit (sort of :)
mem			res	1				' second operand temporary value
aux			res	1				' auxiliary flag temp
vector			res	1				' opcode vector
lmm_pc								' LMM program counter

			fit	$1f0
			org	$1f0
__par			res	1				' read only
__cnt			res	1				' read only
__ina			res	1				' read only
__inb			res	1				' read only
__outa			res	1				' read/write
__outb			res	1				' read/write
__dira			res	1				' read/write - leave this alone
__dirb			res	1				' read/write - leave this alone
__ctra			res	1				' read/write
__ctrb			res	1				' read/write
__frqa			res	1				' read/write
__frqb			res	1				' read/write
__phsa			res	1				' read/write
__phsb			res	1				' read/write
__vcfg			res	1				' read/write - leave this alone
__vscl			res	1				' read/write

PC_reg			word	0
P_reg			byte	0
A_reg			byte	0
X_reg			byte	0
Y_reg			byte	0
S_reg			byte	0
opcode_reg		byte	0, 0, 0, 0, 0

#ifdef INTERRUPTS
'***************************************************************************************************************
'
' Handle interrupts: NMI and IRQ
'
do_interrupt
			mov	t1, t1			wc	' test for negative value
		if_c	add	lmm_pc, #4*(:nmi - $ - 1)
:irq			test	PSW, #II		wz	' interrupts inhibited?
		if_nz	jmp	lmm_ret				' yes, just return
			
			mov	ea, PC
			call	#rd_byte			' dummy read
			call	#push_pc
			mov	alu, PSW
			andn	alu, #BF			' clear break flag
			call	#push_alu
			or	PSW, #II
			mov	PC, low_word
			sub	PC, #1				' PC = $fffe
			call	#rd_opword			' get IRQ vector
			mov	PC, alu
			jmp	lmm_ret				' continue
:nmi
			wrlong	i_nop, irq_ptr			' clear NMI request
			mov	ea, PC
			call	#rd_byte			' dummy read
			call	#push_pc
			mov	alu, PSW
			andn	alu, #BF			' clear break flag
			call	#push_alu
			or	PSW, #II
			mov	PC, low_word
			sub	PC, #5				' PC = $fffa
			call	#rd_opword			' get NMI vector
			mov	PC, alu
			jmp	lmm_ret				' continue
#endif INTERRUPTS

'***************************************************************************************************************
'
' Break
'
lmm_brk
			call	#rd_opcode
			call	#push_pc
			mov	alu, PSW
			or	alu, #BF
			call	#push_alu
			or	PSW, #II
			mov	PC, low_word
			sub	PC, #1				' PC = $fffe
			call	#rd_opword
			jmp	#fetch

'***************************************************************************************************************
'
' Trace call
'
lmm_trace
			mov	t1, context_ptr
			wrword	PC, t1				' store PC
			add	t1, #2
			wrbyte	PSW, t1				' store PSW
			add	t1, #1
			wrbyte	A, t1				' store accumulator
			add	t1, #1
			wrbyte	X, t1				' store X register
			add	t1, #1
			wrbyte	Y, t1				' store Y register
			add	t1, #1
			wrbyte	S, t1				' store stack pointer
			add	t1, #1
			call	#rd_opcode
			wrbyte	alu, t1				' store opcode byte #1
			add	t1, #1
			call	#rd_opcode
			wrbyte	alu, t1				' store opcode byte #2
			add	t1, #1
			call	#rd_opcode
			wrbyte	alu, t1				' store opcode byte #3
			sub	PC, #3
			and	PC, low_word
			mov	t1, #io_cmd_break		' break command
			shl	t1, #16
			or	t1, context_ptr			' addr = context_ptr
			shl	t1, #8
			wrlong	t1, iocmd_ptr
:wait			rdlong	t1, iocmd_ptr
			shr	t1, #24			wz, nr
		if_nz	sub	lmm_pc, #4*($ + 1 - :wait)
			jmp	lmm_ret				' return to caller

'***************************************************************************************************************
'
' Illegal opcode
'
lmm_ill
			mov	t1, context_ptr
			sub	PC, #1
			and	PC, low_word
			wrword	PC, t1				' store PC
			add	t1, #2
			wrbyte	PSW, t1				' store PSW
			add	t1, #1
			wrbyte	A, t1				' store A
			add	t1, #1
			wrbyte	X, t1				' store X
			add	t1, #1
			wrbyte	Y, t1				' store Y
			add	t1, #1
			wrbyte	S, t1				' store S
			add	t1, #1
			call	#rd_opcode
			wrbyte	alu, t1				' store opcode byte #1
			add	t1, #1
			call	#rd_opcode
			wrbyte	alu, t1				' store opcode byte #2
			add	t1, #1
			call	#rd_opcode
			wrbyte	alu, t1				' store opcode byte #3
			sub	PC, #2
			and	PC, low_word
			mov	t1, #io_cmd_break		' break command
			shl	t1, #16
			or	t1, context_ptr			' addr = context_ptr
			shl	t1, #8
			wrlong	t1, iocmd_ptr
:wait			rdlong	t1, iocmd_ptr
			shr	t1, #24			wz, nr
		if_nz	sub	lmm_pc, #4*($ + 1 - :wait)
			jmp	#fetch_0			' fetch next instruction

'***************************************************************************************************************
'
' Add with carry
'
lmm_adc
			mov	mem, alu
			test	PSW, #DF		wz	' decimal mode?
		if_nz	add	lmm_pc, #4*(:decimal - $ - 1)
			test	PSW, #CF		wc
			mov	alu, A
			mov	aux, A
			addx	alu, mem			' add memory operand with carry
			test	alu, #$100		wc	' check for carry
			muxc	PSW, #CF | VF			' set C flags and V flag
			xor	aux, alu
			xor	aux, mem
			test	aux, #$80		wc
		if_c	xor	PSW, #VF			' toggle V flag if sign didn't change
			movs	nz_flags_ret, #wb_acc
			jmp	#nz_flags
:decimal	
			test	cflag, PSW		wc
			mov	lo, A
			and	lo, #$0f
			mov	t1, mem
			and	t1, #$0f
			addx	lo, t1
			mov	hi, A
			and	hi, #$f0
			mov	t1, mem
			and	t1, #$f0
			add	hi, t1
			andn	PSW, #(NF|VF|ZF|CF)
			cmp	lo, #$0a		wc
		if_nc	add	lo, #$06
		if_nc	add	hi, #$10
			mov	aux, A
			xor	aux, hi
			mov	t1, A
			xor	t1, mem
			andn	aux, t1
			test	aux, #$80		wc
			muxc	PSW, #VF
			cmp	hi, #$a0		wc
		if_nc	add	hi, #$60
			test	hi, #$100		wc
			muxc	PSW, #CF
			test	hi, #$80		wc
			muxc	PSW, #NF
			and	hi, #$f0
			and	lo, #$0f
			mov	A, hi
			or	A, lo
			jmp	#fetch

'***************************************************************************************************************
'
' Subtract with carry
'
lmm_sbc
			mov	mem, alu
			test	PSW, #DF		wz	' decimal mode?
		if_nz	add	lmm_pc, #4*(:decimal - $ - 1)
			testn	cflag, PSW		wc	' get !C flag into carry
			mov	alu, A
			mov	aux, A
			subx	alu, mem			' subtract memory operand with borrow
			test	alu, #$100		wc	' check for carry
			muxnc	PSW, #CF			' set C flag if no carry was generated
			muxc	PSW, #VF			' set V flag if carry was generated
			xor	aux, alu
			xor	aux, mem
			test	aux, #$80		wc
		if_c	xor	PSW, #VF
			movs	nz_flags_ret, #wb_acc
			jmp	#nz_flags
:decimal	
			testn	cflag, PSW		wc	' !carry flag
			mov	lo, A
			and	lo, #$0f
			mov	t1, mem
			and	t1, #$0f
			subx	lo, t1				' lo = (A & $0f) - (mem & $0f) - c
			mov	hi, A
			and	hi, #$f0
			mov	t1, mem
			and	t1, #$f0
			sub	hi, t1				' hi = (A & $f0) - (mem & $f0)
			andn	PSW, #(NF|VF|ZF|CF)
			test	lo, #$10		wc
		if_c	sub	lo, #$06
		if_c	sub	hi, #$10
			mov	aux, A
			xor	aux, hi
			mov	t1, A
			xor	t1, mem
			and	aux, t1
			test	aux, #$80		wc
			muxc	PSW, #VF
			test	hi, #$100		wc
		if_c	sub	hi, #$60
			test	hi, #$100		wc
			muxc	PSW, #CF
			test	hi, #$80		wc
			muxc	PSW, #NF
			and	hi, #$f0
			and	lo, #$0f
			mov	A, hi
			or	A, lo
			jmp	#fetch

'***************************************************************************************************************
'
' Jump to subroutine
'
lmm_jsr
			call	#rd_opcode
			mov	t1, alu
			call	#push_pc
			call	#rd_opcode
			shl	alu, #8
			or	alu, t1
			mov	PC, alu
			jmp	#fetch

'***************************************************************************************************************
'
' Return from subroutine
'
lmm_rts
			call	#rd_opcode			' for emulation accuracy's sake: read memory (PC++)
			call	#pull_alu
			mov	PC, alu
			call	#pull_alu
			shl	alu, #8
			or	PC, alu
			movs	rd_opcode_ret, #fetch
			jmp	#rd_opcode			' read memory (PC++)

'***************************************************************************************************************
'
' Return from interrupt subroutine
'
lmm_rti
			call	#rd_opcode			' for emulation accuracy's sake: read memory (PC++)
			call	#rd_stack
			call	#pull_alu
			mov	PSW, alu
			or	PSW, #(TF | BF)
			call	#pull_alu
			mov	PC, alu
			call	#pull_alu
			shl	alu, #8
			or	PC, alu
#ifdef INTERRUPTS
			test	PSW, #II		wz
		if_z	jmp	#fetch_0			' no interrupt immediately after RTI
#endif
			jmp	#fetch

'***************************************************************************************************************
'
' Jump to indirect address
'
lmm_jmp_ind
			call	#rd_opword
			mov	ea, alu
			call	#rd_byte
			mov	t1, ea
			add	t1, #1
			and	t1, #$100
			add	ea, #1
			sub	ea, t1
			mov	PC, alu
			call	#rd_byte
			shl	alu, #8
			or	PC, alu
			jmp	#fetch


'***************************************************************************************************************
'
' (6510) RRA: rotate right and add with carry
' C -> [7][6][5][4][3][2][1][0] -> new C
'
lmm_rra
			test	PSW, #CF		wz
			shr	alu, #1			wc
			muxc	PSW, #CF
			muxnz	alu, #%10000000
			mov	mem, alu
			test	PSW, #DF		wz	' decimal mode?
		if_nz	add	lmm_pc, #4*(:decimal - $ - 1)
			test	PSW, #CF		wc
			mov	alu, A
			mov	aux, A
			addx	alu, mem			' add memory operand with carry
			test	alu, #$100		wc	' check for carry
			muxc	PSW, #CF | VF			' set C flags and V flag
			xor	aux, alu
			xor	aux, mem
			test	aux, #$80		wc
		if_c	xor	PSW, #VF			' toggle V flag if sign didn't change
			movs	nz_flags_ret, #wb_acc
			jmp	#nz_flags
:decimal	
			test	cflag, PSW		wc
			mov	lo, A
			and	lo, #$0f
			mov	t1, mem
			and	t1, #$0f
			addx	lo, t1
			mov	hi, A
			and	hi, #$f0
			mov	t1, mem
			and	t1, #$f0
			add	hi, t1
			andn	PSW, #(NF|VF|ZF|CF)
			cmp	lo, #$0a		wc
		if_nc	add	lo, #$06
		if_nc	add	hi, #$10
			mov	aux, A
			xor	aux, hi
			mov	t1, A
			xor	t1, mem
			andn	aux, t1
			test	aux, #$80		wc
			muxc	PSW, #VF
			cmp	hi, #$a0		wc
		if_nc	add	hi, #$60
			test	hi, #$100		wc
			muxc	PSW, #CF
			test	hi, #$80		wc
			muxc	PSW, #NF
			and	hi, #$f0
			and	lo, #$0f
			mov	A, hi
			or	A, lo
			jmp	#fetch

'***************************************************************************************************************
'
' (6510) ISB: Increment and subtract with carry
'
lmm_isb
			mov	mem, alu
			call	#wr_byte			' pre operation write back
			add	mem, #1
			and	mem, #$ff
			test	PSW, #DF		wz	' decimal mode?
		if_nz	add	lmm_pc, #4*(:decimal - $ - 1)
			testn	cflag, PSW		wc	' get !C flag into carry
			mov	alu, A
			mov	aux, A
			subx	alu, mem			' subtract memory operand with borrow
			test	alu, #$100		wc	' check for carry
			muxnc	PSW, #CF			' set C flag if no carry was generated
			muxc	PSW, #VF			' set V flag if carry was generated
			xor	aux, alu
			xor	aux, mem
			test	aux, #$80		wc
		if_c	xor	PSW, #VF
			movs	nz_flags_ret, #wb_acc
			jmp	#nz_flags
:decimal	
			testn	cflag, PSW		wc	' !carry flag
			mov	lo, A
			and	lo, #$0f
			mov	t1, mem
			and	t1, #$0f
			subx	lo, t1				' lo = (A & $0f) - (mem & $0f) - c
			mov	hi, A
			and	hi, #$f0
			mov	t1, mem
			and	t1, #$f0
			sub	hi, t1				' hi = (A & $f0) - (mem & $f0)
			andn	PSW, #(NF|VF|ZF|CF)
			test	lo, #$10		wc
		if_c	sub	lo, #$06
		if_c	sub	hi, #$10
			mov	aux, A
			xor	aux, hi
			mov	t1, A
			xor	t1, mem
			and	aux, t1
			test	aux, #$80		wc
			muxc	PSW, #VF
			test	hi, #$100		wc
		if_c	sub	hi, #$60
			test	hi, #$100		wc
			muxc	PSW, #CF
			test	hi, #$80		wc
			muxc	PSW, #NF
			and	hi, #$f0
			and	lo, #$0f
			mov	A, hi
			or	A, lo
			jmp	#fetch

'***************************************************************************************************************
'
' (6510) RLA: rotate left and logical and accumulator
' new C <- [7][6][5][4][3][2][1][0] <- C
'
lmm_rla
			test	PSW, #CF		wc
			rcl	alu, #1
			test	alu, #$100		wc
			muxc	PSW, #CF
			and	alu, A
			mov	A, alu
			jmp	#wb_ea

'***************************************************************************************************************
'
' (6510) SSH  store stack high
' logical and accumulator with index X, transfer result to S
' logical and result with effective address high + 1
'
lmm_ssh_aby_np
			mov	XY, Y
			call	#calc_abs_xy_np
			mov	S, A
			and	S, X
			mov	aux, ea
			shr	aux, #8
			add	aux, #1
			mov	alu, S
			and	alu, aux
			call	#wr_byte
			jmp	#fetch

'***************************************************************************************************************
'
' (6510) SYH  store index Y and (high + 1)
' logical and index Y with effective address high + 1 and store the result
lmm_syh_abx_np
			mov	XY, X
			call	#calc_abs_xy_np
			mov	aux, ea
			shr	aux, #8
			add	aux, #1
			mov	alu, Y
			and	alu, aux
			call	#wr_byte
			jmp	#fetch

'***************************************************************************************************************
'
' (6510) SXH  store index X and (high + 1)
' logical and index Y with effective address high + 1 and store the result
lmm_sxh_aby_np
			mov	XY, Y
			call	#calc_abs_xy_np
			mov	aux, ea
			shr	aux, #8
			add	aux, #1
			mov	alu, Y
			and	alu, aux
			call	#wr_byte
			jmp	#fetch

'***************************************************************************************************************
'
' (6510) SAH  store accumulator and index X and ea high + 1
' result = accumulator and index X and effective address high + 1
'
lmm_sah_idy_np
			call	#calc_idy_np
			mov	aux, ea
			shr	aux, #8
			add	aux, #1
			mov	alu, A
			and	alu, X
			and	alu, aux
			jmp	#wb_ea

'***************************************************************************************************************
'
' (6510) SAH  store accumulator and index X and ea high + 1
' result = accumulator and index X and effective address high + 1
'
lmm_sah_aby_np
			mov	XY, Y
			call	#calc_abs_xy_np
			mov	aux, ea
			shr	aux, #8
			add	aux, #1
			mov	alu, A
			and	alu, X
			and	alu, aux
			jmp	#wb_ea

'***************************************************************************************************************
'
' (6510) AST  and stack; transfer to accumulator and index X
' logical and stack (LSB) with data, transfer result to S
' transfer result to accumulator and index X also
lmm_ast_aby_p
			mov	XY, Y
			call	#calc_abs_xy_p
			and	alu, S
			mov	A, alu
			mov	X, alu
			mov	S, alu
			call	#nz_flags
			jmp	#fetch

'***************************************************************************************************************
'
' (6510) ARR logical and, rotate right
'
lmm_arr_imm
			call	#rd_opcode
			test	PSW, #DF		wz
		if_z	add	lmm_pc, #4*(:bin - $ - 1)
			and	alu, A
			mov	aux, alu
			mov	hi, alu
			and	hi, #$f0
			mov	lo, alu
			and	lo, #$0f
			test	PSW, #CF		wc
			shr	alu, #1
			muxc	alu, #$80
			muxc	PSW, #NF
			and	alu, #$ff		wz
			muxz	PSW, #ZF
			xor	aux, alu
			test	aux, #%01000000		wz
			muxnz	PSW, #VF
			mov	aux, lo
			and	aux, #$01
			add	lo, aux
			cmp	lo, #$06		wc
		if_nc	add	alu, #$06
		if_nc	and	alu, #$0f
		if_nc	or	alu, hi
			mov	aux, hi
			and	aux, #$10
			add	hi, aux
			cmp	hi, #$60		wc
		if_nc	or	PSW, #CF
		if_nc	add	alu, #$60
		if_nc	and	alu, #$ff
		if_c	andn	PSW, #CF
			jmp	#wb_acc
:bin
			and	alu, A
			test	PSW, #CF		wz
			shr	alu, #1			wc
			muxc	PSW, #CF
			muxnz	alu, #%10000000
			call	#nz_flags
			test	alu, #%01000000		wz
			muxnz	PSW, #CF
			andn	PSW, #VF
			test	alu, #%00100000		wc
	if_z_and_c	or	PSW, #VF
	if_nz_and_nc	or	PSW, #VF
			jmp	#wb_acc

'***************************************************************************************************************
'
' (6510) ASX logical and X w/ A, subtract data from X
'
lmm_asx_imm
			call	#rd_opcode
			and	X, A
			cmp	X, alu			wc
			muxnc	PSW, #CF
			mov	alu, X
			call	#nz_flags
			jmp	#fetch

'***************************************************************************************************************
'
' (6510) AXA transfer index X to accumulator, logical and depends on the data of the DMA
' device (videochip) fetched between opcode read and operand read
'
lmm_axa_imm
			call	#rd_opcode
			or	A, #$ee
			and	A, X
			and	A, alu
			mov	alu, A
			call	#nz_flags
			jmp	#fetch

'***************************************************************************************************************
'
' (6510) OAL load accumulator and index X
'
lmm_oal_imm
			call	#rd_opcode
			or	A, #$ee
			and	alu, A
			mov	A, alu
			mov	X, alu
			call	#nz_flags
			jmp	#fetch

opcodes		long
		{00}	long	lmm         | lmm_brk_  << 9						'' 7 BRK
		{01}	long	rd_idx	    | alu_ora   << 9						'' 6 ORA IDX
		{02}	long	lmm         | lmm_ill_  << 9						'' 1 ILL
		{03}	long	rd_idx      | wb_ea     << 9    | alu_slo   << 18 | wb_ea  << 27	'' 7 SLO IDX
		{04}	long	rd_zpg      | m_nop     << 9						'' 3 NOP ZPG
		{05}	long	rd_zpg      | alu_ora   << 9						'' 3 ORA ZPG
		{06}	long	rd_zpg      | wb_ea     << 9    | alu_asl   << 18 | wb_ea  << 27	'' 5 ASL ZPG
		{07}	long	rd_zpg      | wb_ea     << 9    | alu_slo   << 18 | wb_ea  << 27	'' 5 SLO ZPG
		{08}	long	rd_dummy    | m_php     << 9						'' 3 PHP
		{09}	long	rd_imm      | alu_ora   << 9						'' 2 ORA IMM
		{0a}	long	rd_dummy    | rd_acc    << 9    | alu_asl   << 18 | wb_acc << 27	'' 2 ASL A
		{0b}	long	rd_imm      | alu_anc   << 9						'' 2 ANC IMM
		{0c}	long	rd_abs      | m_nop     << 9						'' 4 NOP ABS
		{0d}	long	rd_abs      | alu_ora   << 9						'' 4 ORA ABS
		{0e}	long	rd_abs      | wb_ea     << 9    | alu_asl   << 18 | wb_ea  << 27	'' 6 ASL ABS
		{0f}	long	rd_abs      | wb_ea     << 9    | alu_slo   << 18 | wb_ea  << 27	'' 6 SLO ABS
		{10}	long	bra_clr     | NF        << 9						'' 2-4 BPL REL
		{11}	long	rd_idy_p    | alu_ora   << 9						'' 5 ORA IDY [*]
		{02}	long	lmm         | lmm_ill_  << 9						'' 1 ILL
		{13}	long	rd_idy_np   | wb_ea     << 9    | alu_slo   << 18 | wb_ea  << 27	'' 7 SLO IDY
		{14}	long	rd_zpx      | m_nop     << 9						'' 4 NOP ZPX
		{15}	long	rd_zpx      | alu_ora   << 9						'' 4 ORA ZPX
		{16}	long	rd_zpx      | wb_ea     << 9    | alu_asl   << 18 | wb_ea  << 27	'' 6 ASL ZPX
		{17}	long	rd_zpx      | wb_ea     << 9    | alu_slo   << 18 | wb_ea  << 27	'' 6 SLO ZPX
		{18}	long	rd_dummy    | clr_flag  << 9    | CF        << 18			'' 2 CLC
		{19}	long	rd_aby_p    | alu_ora   << 9 						'' 4 ORA ABY [*]
		{1a}	long	rd_dummy    | m_nop     << 9						'' 2 NOP
		{1b}	long	rd_aby_np   | wb_ea     << 9    | alu_slo   << 18 | wb_ea  << 27	'' 7 SLO ABY
		{1c}	long	rd_abx_p    | m_nop     << 9						'' 4 NOP ABX [*]
		{1d}	long	rd_abx_p    | alu_ora   << 9						'' 4 ORA ABX [*]
		{1e}	long	rd_abx_np   | wb_ea     << 9    | alu_asl   << 18 | wb_ea  << 27	'' 7 ASL ABX
		{1f}	long	rd_abx_np   | wb_ea     << 9    | alu_slo   << 18 | wb_ea  << 27	'' 7 SLO ABX
		{20}	long	lmm         | lmm_jsr_  << 9						'' 6 JSR
		{21}	long	rd_idx      | alu_and   << 9						'' 6 AND IDX
		{02}	long	lmm         | lmm_ill_  << 9						'' 1 ILL
		{23}	long	rd_idx      | lmm       << 9    | lmm_rla_  << 18			'' 7 RLA IDX
		{24}	long	rd_zpg      | alu_bit   << 9						'' 3 BIT ZPG
		{25}	long	rd_zpg      | alu_and   << 9						'' 3 AND ZPG
		{26}	long	rd_zpg      | wb_ea     << 9    | alu_rol   << 18 | wb_ea  << 27	'' 5 ROL ZPG
		{27}	long	rd_zpg      | lmm       << 9    | lmm_rla_  << 18			'' 5 RLA ZPG
		{28}	long	rd_dummy    | m_plp     << 9						'' 4 PLP
		{29}	long	rd_imm      | alu_and   << 9						'' 2 AND IMM
		{2a}	long	rd_dummy    | rd_acc    << 9    | alu_rol   << 18 | wb_acc << 27	'' 2 ROL A
		{2b}	long	rd_imm      | alu_anc   << 9						'' 2 ANC IMM
		{2c}	long	rd_abs      | alu_bit   << 9						'' 4 BIT ABS
		{2d}	long	rd_abs      | alu_and   << 9						'' 4 AND ABS
		{2e}	long	rd_abs      | wb_ea     << 9    | alu_rol   << 18 | wb_ea  << 27	'' 6 ROL ABS
		{2f}	long	rd_abs      | lmm       << 9    | lmm_rla_  << 18			'' 6 RLA ABS
		{30}	long	bra_set     | NF        << 9						'' 2-4 BMI REL
		{31}	long	rd_idy_p    | alu_and   << 9						'' 5 AND IDY [*]
		{02}	long	lmm         | lmm_ill_  << 9						'' 1 ILL
		{33}	long	rd_idy_np   | lmm       << 9    | lmm_rla_  << 18			'' 7 RLA IDY
		{34}	long	rd_zpx      | m_nop     << 9						'' 4 NOP ZPX
		{35}	long	rd_zpx      | alu_and   << 9						'' 4 AND ZPX
		{36}	long	rd_zpx      | wb_ea     << 9    | alu_rol   << 18 | wb_ea  << 27	'' 6 ROL ZPX
		{37}	long	rd_zpx      | lmm       << 9    | lmm_rla_  << 18			'' 6 RLA ZPX
		{38}	long	rd_dummy    | set_flag  << 9    | CF        << 18			'' 2 SEC
		{39}	long	rd_aby_p    | alu_and   << 9						'' 4 AND ABY [*]
		{3a}	long	rd_dummy    | m_nop     << 9						'' 2 NOP
		{3b}	long	rd_aby_np   | lmm       << 9    | lmm_rla_  << 18			'' 7 RLA ABY
		{3c}	long	rd_abx_p    | m_nop     << 9						'' 4 NOP ABX [*]
		{3d}	long	rd_abx_p    | alu_and   << 9						'' 4 AND ABX [*]
		{3e}	long	rd_abx_np   | wb_ea     << 9    | alu_rol   << 18 | wb_ea	<< 27	'' 7 ROL ABX
		{3f}	long	rd_abx_np   | lmm       << 9    | lmm_rla_  << 18			'' 7 RLA ABX
		{40}	long	lmm         | lmm_rti_  << 9						'' 6 RTI
		{41}	long	rd_idx      | alu_eor   << 9						'' 6 EOR IDX
		{02}	long	lmm         | lmm_ill_  << 9						'' 1 ILL
		{43}	long	rd_idx      | wb_ea     << 9    | alu_sre   << 18 | wb_ea  << 27	'' 7 SRE IDX
		{44}	long	rd_zpg      | m_nop     << 9						'' 3 NOP ZPG
		{45}	long	rd_zpg      | alu_eor   << 9						'' 3 EOR ZPG
		{46}	long	rd_zpg      | wb_ea     << 9    | alu_lsr   << 18 | wb_ea  << 27	'' 5 LSR ZPG
		{47}	long	rd_zpg      | wb_ea     << 9    | alu_sre   << 18 | wb_ea  << 27	'' 5 SRE ZPG
		{48}	long	rd_dummy    | m_pha     << 9						'' 3 PHA
		{49}	long	rd_imm      | alu_eor   << 9						'' 2 EOR IMM
		{4a}	long	rd_dummy    | rd_acc    << 9    | alu_lsr   << 18 | wb_acc << 27	'' 2 LSR A
		{4b}	long	rd_imm      | alu_asr   << 9    | wb_acc    << 18			'' 2 ASR IMM
		{4c}	long	ea_abs      | m_jmp     << 9						'' 3 JMP ABS
		{4d}	long	rd_abs      | alu_eor   << 9						'' 4 EOR ABS
		{4e}	long	rd_abs      | wb_ea     << 9    | alu_lsr   << 18 | wb_ea  << 27	'' 6 LSR ABS
		{4f}	long	rd_abs      | wb_ea     << 9    | alu_sre   << 18 | wb_ea  << 27	'' 6 SRE ABS
		{50}	long	bra_clr     | VF        << 9						'' 2-4 BVC REL
		{51}	long	rd_idy_p    | alu_eor   << 9						'' 5 EOR IDY [*]
		{52}	long	lmm         | lmm_ill_  << 9						'' 1 ILL
		{53}	long	rd_idy_np   | wb_ea     << 9    | alu_sre   << 18 | wb_ea  << 27	'' 7 SRE IDY
		{54}	long	rd_zpx      | m_nop     << 9						'' 4 NOP ZPX
		{55}	long	rd_zpx      | alu_eor   << 9						'' 4 EOR ZPX
		{56}	long	rd_zpx      | wb_ea     << 9    | alu_lsr   << 18 | wb_ea  << 27	'' 6 LSR ZPX
		{57}	long	rd_zpx      | wb_ea     << 9    | alu_sre   << 18 | wb_ea  << 27	'' 6 SRE ZPX
		{58}	long	rd_dummy    | clr_flag  << 9    | II        << 18			'' 2 CLI
		{59}	long	rd_aby_p    | alu_eor   << 9						'' 4 EOR ABY [*]
		{5a}	long	rd_dummy    | m_nop     << 9						'' 2 NOP
		{5b}	long	rd_aby_np   | wb_ea     << 9    | alu_sre   << 18 | wb_ea  << 27	'' 7 SRE ABY
		{5c}	long	rd_abx_p    | m_nop     << 9						'' 4 NOP ABX [*]
		{5d}	long	rd_abx_p    | alu_eor   << 9						'' 4 EOR ABX [*]
		{5e}	long	rd_abx_np   | wb_ea     << 9    | alu_lsr   << 18 | wb_ea  << 27	'' 7 LSR ABX
		{5f}	long	rd_abx_np   | wb_ea     << 9    | alu_sre   << 18 | wb_ea  << 27	'' 7 SRE ABX
		{60}	long	lmm         | lmm_rts_  << 9						'' 6 RTS
		{61}	long	rd_idx      | lmm       << 9    | lmm_adc_  << 18			'' 6 ADC IDX
		{62}	long	lmm         | lmm_ill_  << 9						'' 1 ILL
		{63}	long	rd_idx      | lmm       << 9    | lmm_rra_  << 18			'' 7 RRA IDX
		{64}	long	rd_zpg      | m_nop     << 9						'' 3 NOP ZPG
		{65}	long	rd_zpg      | lmm       << 9    | lmm_adc_  << 18			'' 3 ADC ZPG
		{66}	long	rd_zpg      | wb_ea     << 9    | alu_ror   << 18 | wb_ea  << 27	'' 5 ROR ZPG
		{67}	long	rd_zpg      | lmm       << 9    | lmm_rra_  << 18			'' 5 RRA ZPG
		{68}	long	rd_dummy    | m_pla     << 9						'' 4 PLA
		{69}	long	rd_imm      | lmm       << 9    | lmm_adc_  << 18			'' 2 ADC IMM
		{6a}	long	rd_dummy    | rd_acc    << 9    | alu_ror   << 18 | wb_acc << 27	'' 2 ROR A
		{6b}	long	lmm         | lmm_arr_imm_ << 9						'' 2 ARR IMM
		{6c}	long	lmm         | lmm_jmp_ind_ << 9						'' 5 JMP IND
		{6d}	long	rd_abs      | lmm       << 9    | lmm_adc_  << 18			'' 4 ADC ABS
		{6e}	long	rd_abs      | wb_ea     << 9    | alu_ror   << 18 | wb_ea  << 27	'' 6 ROR ABS
		{6f}	long	rd_abs      | lmm       << 9    | lmm_rra_  << 18			'' 6 RRA ABS
		{70}	long	bra_set     | VF        << 9						'' 2-4 BVS REL
		{71}	long	rd_idy_p    | lmm       << 9    | lmm_adc_  << 18			'' 5 ADC IDY [*]
		{72}	long	lmm         | lmm_ill_  << 9						'' 1 ILL
		{73}	long	rd_idy_np   | lmm       << 9    | lmm_rra_  << 18			'' 7 RRA IDY
		{74}	long	rd_zpx      | m_nop     << 9						'' 4 NOP ZPX
		{75}	long	rd_zpx      | lmm       << 9    | lmm_adc_  << 18			'' 4 ADC ZPX
		{76}	long	rd_zpx      | wb_ea     << 9    | alu_ror   << 18 | wb_ea  << 27	'' 6 ROR ZPX
		{77}	long	rd_zpx      | lmm       << 9    | lmm_rra_  << 18			'' 6 RRA ZPX
		{78}	long	rd_dummy    | set_flag  << 9    | II        << 18			'' 2 SEI
		{79}	long	rd_aby_p    | lmm       << 9    | lmm_adc_  << 18			'' 4 ADC ABY [*]
		{7a}	long	rd_dummy    | m_nop     << 9						'' 2 NOP
		{7b}	long	rd_aby_np   | lmm       << 9    | lmm_rra_  << 18			'' 7 RRA ABY
		{7c}	long	rd_abx_p    | m_nop     << 9						'' 4 NOP ABX [*]
		{7d}	long	rd_abx_p    | lmm       << 9    | lmm_adc_  << 18 			'' 4 ADC ABX [*]
		{7e}	long	rd_abx_np   | wb_ea     << 9    | alu_ror   << 18 | wb_ea  << 27	'' 7 ROR ABX
		{7f}	long	rd_abx_np   | lmm       << 9    | lmm_rra_  << 18			'' 7 RRA ABX
		{80}	long	rd_imm      | m_nop     << 9						'' 2 NOP IMM
		{81}	long	m_sta       | wr_idx    << 9						'' 6 STA IDX
		{82}	long	rd_imm      | m_nop     << 9						'' 2 NOP IMM
		{83}	long	m_sax       | wr_idx    << 9						'' 6 SAX IDX
		{84}	long	m_sty       | wr_zpg    << 9						'' 3 STY ZPG
		{85}	long	m_sta       | wr_zpg    << 9						'' 3 STA ZPG
		{86}	long	m_stx       | wr_zpg    << 9						'' 3 STX ZPG
		{87}	long	m_sax       | wr_zpg    << 9						'' 3 SAX ZPG
		{88}	long	rd_dummy    | m_dey     << 9						'' 2 DEY
		{89}	long	rd_imm      | m_nop     << 9						'' 2 NOP IMM
		{8a}	long	rd_dummy    | m_txa     << 9						'' 2 TXA
		{8b}	long	rd_imm      | lmm       << 9    | lmm_axa_imm_ << 18			'' 2 AXA IMM
		{8c}	long	m_sty       | wr_abs    << 9						'' 4 STY ABS
		{8d}	long	m_sta       | wr_abs    << 9						'' 4 STA ABS
		{8e}	long	m_stx       | wr_abs    << 9						'' 4 STX ABS
		{8f}	long	m_sax       | wr_abs    << 9						'' 4 SAX ABS
		{90}	long	bra_clr     | CF        << 9						'' 2-4 BCC REL
		{91}	long	m_sta       | wr_idy_np << 9						'' 6 STA IDY
		{92}	long	lmm         | lmm_ill_  << 9						'' 1 ILL
		{93}	long	lmm         | lmm_sah_idy_np_ << 9					'' 5 SAH IDY
		{94}	long	m_sty       | wr_zpx    << 9						'' 4 STY ZPX
		{95}	long	m_sta       | wr_zpx    << 9						'' 4 STA ZPX
		{96}	long	m_stx       | wr_zpy    << 9						'' 4 STX ZPY
		{97}	long	m_sax       | wr_zpy    << 9						'' 4 SAX ZPY
		{98}	long	rd_dummy    | m_tya     << 9						'' 2 TYA
		{99}	long	m_sta       | wr_aby_np << 9						'' 5 STA ABY
		{9a}	long	rd_dummy    | m_txs     << 9						'' 2 TXS
		{9b}	long	lmm         | lmm_ssh_aby_np_ << 9					'' 5 SSH ABY
		{9c}	long	lmm         | lmm_syh_abx_np_ << 9					'' 5 SYH ABX
		{9d}	long	m_sta       | wr_abx_np << 9						'' 5 STA ABX
		{9e}	long	lmm         | lmm_sxh_aby_np_ << 9					'' 5 SXH ABY
		{9f}	long	lmm         | lmm_sah_aby_np_ << 9					'' 5 SAH ABY
		{a0}	long	rd_imm      | m_ldy     << 9						'' 2 LDY IMM
		{a1}	long	rd_idx      | m_lda     << 9						'' 6 LDA IDX
		{a2}	long	rd_imm      | m_ldx     << 9						'' 2 LDX IMM
		{a3}	long	rd_idx      | m_lax     << 9						'' 6 LAX IDX
		{a4}	long	rd_zpg      | m_ldy     << 9						'' 3 LDY ZPG
		{a5}	long	rd_zpg      | m_lda     << 9						'' 3 LDA ZPG
		{a6}	long	rd_zpg      | m_ldx     << 9						'' 3 LDX ZPG
		{a7}	long	rd_zpg      | m_lax     << 9						'' 3 LAX ZPG
		{a8}	long	rd_dummy    | m_tay     << 9						'' 2 TAY
		{a9}	long	rd_imm      | m_lda     << 9						'' 2 LDA IMM
		{aa}	long	rd_dummy    | m_tax     << 9						'' 2 TAX
		{ab}	long	lmm         | lmm_oal_imm_ << 9						'' 2 OAL IMM
		{ac}	long	rd_abs      | m_ldy     << 9						'' 4 LDY ABS
		{ad}	long	rd_abs      | m_lda     << 9						'' 4 LDA ABS
		{ae}	long	rd_abs      | m_ldx     << 9						'' 4 LDX ABS
		{af}	long	rd_abs      | m_lax     << 9						'' 4 LAX ABS
		{b0}	long	bra_set     | CF        << 9						'' 2-4 BCS REL
		{b1}	long	rd_idy_p    | m_lda     << 9						'' 5 LDA IDY [*]
		{b2}	long	lmm         | lmm_ill_  << 9						'' 1 ILL
		{b3}	long	rd_idy_p    | m_lax     << 9						'' 5 LAX IDY [*]
		{b4}	long	rd_zpx      | m_ldy     << 9						'' 4 LDY ZPX
		{b5}	long	rd_zpx      | m_lda     << 9						'' 4 LDA ZPX
		{b6}	long	rd_zpy      | m_ldx     << 9						'' 4 LDX ZPY
		{b7}	long	rd_zpy      | m_lax     << 9						'' 4 LAX ZPY
		{b8}	long	rd_dummy    | clr_flag  << 9    | VF        << 18			'' 2 CLV
		{b9}	long	rd_aby_p    | m_lda     << 9						'' 4 LDA ABY [*]
		{ba}	long	rd_dummy    | m_tsx     << 9						'' 2 TSX
		{bb}	long	lmm         | lmm_ast_aby_p_ << 9					'' 4 AST ABY [*]
		{bc}	long	rd_abx_p    | m_ldy     << 9						'' 4 LDY ABX [*]
		{bd}	long	rd_abx_p    | m_lda     << 9						'' 4 LDA ABX [*]
		{be}	long	rd_aby_p    | m_ldx     << 9						'' 4 LDX ABY [*]
		{bf}	long	rd_aby_p    | m_lax     << 9						'' 4 LAX ABY [*]
		{c0}	long	rd_imm      | m_cpy     << 9						'' 2 CPY IMM
		{c1}	long	rd_idx      | alu_cmp   << 9						'' 6 CMP IDX
		{c2}	long	rd_imm      | m_nop     << 9						'' 2 NOP IMM
		{c3}	long	rd_idx      | wb_ea     << 9    | alu_dcp   << 18 | wb_ea  << 27	'' 7 DCP IDX
		{c4}	long	rd_zpg      | m_cpy     << 9						'' 3 CPY ZPG
		{c5}	long	rd_zpg      | alu_cmp   << 9						'' 3 CMP ZPG
		{c6}	long	rd_zpg      | wb_ea     << 9    | alu_dec   << 18 | wb_ea  << 27	'' 5 DEC ZPG
		{c7}	long	rd_zpg      | wb_ea     << 9    | alu_dcp   << 18 | wb_ea  << 27	'' 5 DCP ZPG
		{c8}	long	rd_dummy    | m_iny     << 9						'' 2 INY
		{c9}	long	rd_imm      | alu_cmp   << 9						'' 2 CMP IMM
		{ca}	long	rd_dummy    | m_dex     << 9						'' 2 DEX
		{cb}	long	rd_imm      | lmm       << 9    | lmm_asx_imm_ << 18			'' 2 ASX IMM
		{cc}	long	rd_abs      | m_cpy     << 9						'' 4 CPY ABS
		{cd}	long	rd_abs      | alu_cmp   << 9						'' 4 CMP ABS
		{ce}	long	rd_abs      | wb_ea     << 9    | alu_dec   << 18 | wb_ea  << 27	'' 6 DEC ABS
		{cf}	long	rd_abs      | wb_ea     << 9    | alu_dcp   << 18 | wb_ea  << 27	'' 6 DCP ABS
		{d0}	long	bra_clr     | ZF        << 9						'' 2-4 BNE REL
		{d1}	long	rd_idy_p    | alu_cmp   << 9						'' 5 CMP IDY [*]
		{d2}	long	lmm         | lmm_ill_  << 9						'' 1 ILL
		{d3}	long	rd_idy_np   | wb_ea     << 9    | alu_dcp   << 18 | wb_ea  << 27	'' 7 DCP IDY
		{d4}	long	rd_zpx      | m_nop     << 9						'' 4 NOP ZPX
		{d5}	long	rd_zpx      | alu_cmp   << 9						'' 4 CMP ZPX
		{d6}	long	rd_zpx      | wb_ea     << 9    | alu_dec   << 18 | wb_ea  << 27	'' 6 DEC ZPX
		{d7}	long	rd_zpx      | wb_ea     << 9    | alu_dcp   << 18 | wb_ea  << 27	'' 6 DCP ZPX
		{d8}	long	rd_dummy    | clr_flag  << 9    | DF        << 18			'' 2 CLD
		{d9}	long	rd_aby_p    | alu_cmp   << 9						'' 4 CMP ABY [*]
		{da}	long	rd_dummy    | m_nop     << 9						'' 2 NOP
		{db}	long	rd_aby_np   | wb_ea     << 9    | alu_dcp   << 18 | wb_ea  << 27	'' 7 DCP ABY
		{dc}	long	rd_abx_p    | m_nop     << 9						'' 4 NOP ABX [*]
		{dd}	long	rd_abx_p    | alu_cmp   << 9						'' 4 CMP ABX [*]
		{de}	long	rd_abx_np   | wb_ea     << 9    | alu_dec   << 18 | wb_ea  << 27	'' 7 DEC ABX
		{df}	long	rd_abx_np   | wb_ea     << 9    | alu_dcp   << 18 | wb_ea  << 27	'' 7 DCP ABX
		{e0}	long	rd_imm      | m_cpx     << 9						'' 2 CPX IMM
		{e1}	long	rd_idx      | lmm       << 9    | lmm_sbc_  << 18			'' 6 SBC IDX
		{e2}	long	rd_imm      | m_nop     << 9						'' 2 NOP IMM
		{e3}	long	rd_idx      | lmm       << 9    | lmm_isb_  << 18			'' 7 ISB IDX
		{e4}	long	rd_zpg      | m_cpx     << 9						'' 3 CPX ZPG
		{e5}	long	rd_zpg      | lmm       << 9    | lmm_sbc_  << 18			'' 3 SBC ZPG
		{e6}	long	rd_zpg      | wb_ea     << 9    | alu_inc   << 18 | wb_ea  << 27	'' 5 INC ZPG
		{e7}	long	rd_zpg      | lmm       << 9    | lmm_isb_  << 18			'' 5 ISB ZPG
		{e8}	long	rd_dummy    | m_inx     << 9						'' 2 INX
		{e9}	long	rd_imm      | lmm       << 9    | lmm_sbc_  << 18			'' 2 SBC IMM
		{ea}	long	rd_dummy    | m_nop     << 9						'' 2 NOP
		{eb}	long	rd_imm      | lmm       << 9    | lmm_sbc_  << 18			'' 2 SBC IMM
		{ec}	long	rd_abs      | m_cpx     << 9						'' 4 CPX ABS
		{ed}	long	rd_abs      | lmm       << 9    | lmm_sbc_  << 18			'' 4 SBC ABS
		{ee}	long	rd_abs      | wb_ea     << 9    | alu_inc   << 18 | wb_ea  << 27	'' 6 INC ABS
		{ef}	long	rd_abs      | lmm       << 9    | lmm_isb_  << 18			'' 6 ISB ABS
		{f0}	long	bra_set     | ZF        << 9						'' 2-4 BEQ REL
		{f1}	long	rd_idy_p    | lmm       << 9    | lmm_sbc_  << 18			'' 5 SBC IDY [*]
		{f2}	long	lmm         | lmm_ill_  << 9						'' 1 ILL
		{f3}	long	rd_idy_np   | lmm       << 9    | lmm_isb_  << 18			'' 7 ISB IDY
		{f4}	long	rd_zpx      | m_nop     << 9						'' 4 NOP ZPX
		{f5}	long	rd_zpx      | lmm       << 9    | lmm_sbc_  << 18			'' 4 SBC ZPX
		{f6}	long	rd_zpx      | wb_ea     << 9    | alu_inc   << 18 | wb_ea  << 27	'' 6 INC ZPX
		{f7}	long	rd_zpx      | lmm       << 9    | lmm_isb_  << 18			'' 6 ISB ZPX
		{f8}	long	rd_dummy    | set_flag  << 9    | DF        << 18			'' 2 SED
		{f9}	long	rd_aby_p    | lmm       << 9    | lmm_sbc_  << 18			'' 4 SBC ABY [*]
		{fa}	long	rd_dummy    | m_nop     << 9						'' 2 NOP
		{fb}	long	rd_aby_np   | lmm       << 9    | lmm_isb_  << 18			'' 7 ISB ABY
		{fc}	long	rd_abx_p    | m_nop     << 9						'' 4 NOP ABX [*]
		{fd}	long	rd_abx_p    | lmm       << 9    | lmm_sbc_  << 18			'' 4 SBC ABX [*]
		{fe}	long	rd_abx_np   | wb_ea     << 9    | alu_inc   << 18 | wb_ea  << 27	'' 7 INC ABX
		{ff}	long	rd_abx_np   | lmm       << 9    | lmm_isb_  << 18			'' 7 ISB ABX
''
'' [*] page wrap penalty

lmm_functions	long
		{ 0}	long	@@@lmm_brk
		{ 1}	long	@@@lmm_trace
		{ 2}	long	@@@lmm_ill
		{ 3}	long	@@@lmm_adc
		{ 4}	long	@@@lmm_sbc
		{ 5}	long	@@@lmm_jsr
		{ 6}	long	@@@lmm_rts
		{ 7}	long	@@@lmm_rti
		{ 8}	long	@@@lmm_jmp_ind
		{ 9}	long	@@@lmm_rra
		{10}	long	@@@lmm_isb
		{11}	long	@@@lmm_rla
		{12}	long	@@@lmm_ssh_aby_np
		{13}	long	@@@lmm_syh_abx_np
		{14}	long	@@@lmm_sxh_aby_np
		{15}	long	@@@lmm_sah_idy_np
		{16}	long	@@@lmm_sah_aby_np
		{17}	long	@@@lmm_ast_aby_p
		{18}	long	@@@lmm_arr_imm
		{19}	long	@@@lmm_asx_imm
		{20}	long	@@@lmm_axa_imm
		{21}	long	@@@lmm_oal_imm

#ifdef MICROTAN
rd_memory	' x000   x100   x200   x300   x400   x500   x600   x700   x800   x900   xa00   xb00   xc00   xd00   xe00   xf00
{0xxx}	byte	rd_ram,rd_ram,rd_i_o,rd_i_o,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram
{1xxx}	byte	rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram
{2xxx}	byte	rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram
{3xxx}	byte	rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram
{4xxx}	byte	rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram
{5xxx}	byte	rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram
{6xxx}	byte	rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram
{7xxx}	byte	rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram
{8xxx}	byte	rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram
{9xxx}	byte	rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram
{axxx}	byte	rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram
{bxxx}	byte	rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_ram,rd_i_o,rd_i_o,rd_i_o,rd_i_o
{cxxx}	byte	rd_rom,rd_rom,rd_rom,rd_rom,rd_rom,rd_rom,rd_rom,rd_rom,rd_rom,rd_rom,rd_rom,rd_rom,rd_rom,rd_rom,rd_rom,rd_rom
{dxxx}	byte	rd_rom,rd_rom,rd_rom,rd_rom,rd_rom,rd_rom,rd_rom,rd_rom,rd_rom,rd_rom,rd_rom,rd_rom,rd_rom,rd_rom,rd_rom,rd_rom
{exxx}	byte	rd_rom,rd_rom,rd_rom,rd_rom,rd_rom,rd_rom,rd_rom,rd_rom,rd_rom,rd_rom,rd_rom,rd_rom,rd_rom,rd_rom,rd_rom,rd_rom
{fxxx}	byte	rd_rom,rd_rom,rd_rom,rd_rom,rd_rom,rd_rom,rd_rom,rd_rom,rd_rom,rd_rom,rd_rom,rd_rom,rd_rom,rd_rom,rd_rom,rd_rom

wr_memory	' x000   x100   x200   x300   x400   x500   x600   x700   x800   x900   xa00   xb00   xc00   xd00   xe00   xf00
{0xxx}	byte	wr_ram,wr_ram,wr_i_o,wr_i_o,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram
{1xxx}	byte	wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram
{2xxx}	byte	wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram
{3xxx}	byte	wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram
{4xxx}	byte	wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram
{5xxx}	byte	wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram
{6xxx}	byte	wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram
{7xxx}	byte	wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram
{8xxx}	byte	wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram
{9xxx}	byte	wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram
{axxx}	byte	wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram
{bxxx}	byte	wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_ram,wr_i_o,wr_i_o,wr_i_o,wr_i_o
{cxxx}	byte	wr_rom,wr_rom,wr_rom,wr_rom,wr_rom,wr_rom,wr_rom,wr_rom,wr_rom,wr_rom,wr_rom,wr_rom,wr_rom,wr_rom,wr_rom,wr_rom
{dxxx}	byte	wr_rom,wr_rom,wr_rom,wr_rom,wr_rom,wr_rom,wr_rom,wr_rom,wr_rom,wr_rom,wr_rom,wr_rom,wr_rom,wr_rom,wr_rom,wr_rom
{exxx}	byte	wr_rom,wr_rom,wr_rom,wr_rom,wr_rom,wr_rom,wr_rom,wr_rom,wr_rom,wr_rom,wr_rom,wr_rom,wr_rom,wr_rom,wr_rom,wr_rom
{fxxx}	byte	wr_rom,wr_rom,wr_rom,wr_rom,wr_rom,wr_rom,wr_rom,wr_rom,wr_rom,wr_rom,wr_rom,wr_rom,wr_rom,wr_rom,wr_rom,wr_rom
#endif MICROTRAN
