'******************************************************************************
'*  io.spin - handle input and output requests sent by q6502
'*
'*  (c) Juergen Buchmueller <pullmoll@t-online.de>
'*
'* $Id: io.spin,v 1.2 2010-05-27 19:13:03 pm Exp $
'******************************************************************************
CON

	io_cmd_out		=	$01		' I/O emulator command output
	io_cmd_in		=	$02		' I/O emulator command input
	io_cmd_break		=	$03		' I/O emulator command break
	io_cmd_kbd		=	$04		' I/O emulator command read Microtan keyboard

	baud			=	115200		' baud rate
	SCRATCH_RAM		=	$4000
	FSEL_TOP		=	2
	FSEL_H			=	14

#define FILESELECT

VAR
	long	cog
	long	params[3]
	long	nmi_vector
	long	rst_vector
	long	irq_vector
	long	version
	long	ramsize
	long	rc
	word	topy
	word	boty
	word	ndir
	word	row
	word	y0
	word	speed
	byte	key
	byte	curx
	byte	cury
	byte	getdir
	byte	de[32]					' directory entry
	byte	name[12+1]
	byte	datetime[20+1]
	byte	buff[64]

OBJ
	fat	: "fatfs"
#ifdef TV
	video	: "video"
#endif TV
#ifdef VGA
	video	: "vga"
#endif
	kbd	: "keyboard"
	psg	: "AYcog_V0-22"

PUB start(pins) : okay | i, j, dent, sec
	stop

	iocmd_ptr := @io_command
	irq_ptr := @irq
	screen_ptr := @videoram
	graphics_ptr := @graphicsram
	chr(12)
	str(string("Microtan q6502 emulator",13,10))

	' *************************************************************************************
	' Start the FAT driver and the SD SPI object (must be first)
	'
	okay := \fat.start(byte[pins][2], byte[pins][3], byte[pins][4], byte[pins][5])
	if okay < 0
		failure(string("SPI"),okay)
	else
		success(string("SPI"))
	iobuff_ptr := \fat.get_memory

#ifdef TV
	' *************************************************************************************
	' Start the Microtan TV video object
	'
	params[0] := screen				' 1K video RAM
	params[1] := graphics
	params[2] := byte[pins][9]			' video pin
	okay := video.start(@params)
	if okay < 0
		failure(string("TV"),okay)
		return
#endif TV

#ifdef VGA
	' *************************************************************************************
	' Start the Microtan VGA video object
	'
	okay := video.start(byte[pins][9], @videoram, @graphicsram, @cursor, @sync)
	if okay < 0
		failure(string("VGA"),okay)
		return
	success(string("VGA"))
#endif VGA

	' *************************************************************************************
	' Start the Microtan keyboard driver
	'
	okay := \kbd.startx(byte[pins][6], byte[pins][7], 0, 40)
	if okay < 0
		failure(string("KBD"),okay)
		return
	kbd_head_ptr := kbd.head_ptr
	kbd_tail_ptr := kbd.tail_ptr
	kbd_keys_ptr := kbd.keys_ptr
	success(string("KBD"))

	' *************************************************************************************
	' mount the FAT, i.e. check the format of the SD media
	'
	okay := \fat.mount
	if okay < 0
		failure(string("FAT mount"), okay)
		return -1
	success(string("FAT"))

	longfill(iobuff_ptr, 0, $200/4)
	repeat i from $0000 to $fe00 step $200
		fat.write_xmm_ram(iobuff_ptr, i, $200)

	' *************************************************************************************
	' find the MICROTAN.CHR file and load it into the character generator
	'
	dent := open_file(string("MICROTAN.CHR"))
	if dent => 0
		sec := fat.get_dent_first_sector(@de)
		repeat i from $000 to $600 step $200
			fat.read_sector(sec)
			video.write_chargen(i, iobuff_ptr, $200)
			sec := fat.get_dent_next_sector(@de, sec)
		str(string("Loaded character generator.",13,10))

	' *********************************************************
	' find the MICROTAN.ROM file and load it into the ROM region
	'
	dent := open_file(string("MICROTAN.ROM"))
	if dent < 0
		return -1
	sec := fat.get_dent_first_sector(@de)
	repeat i from $c000 to $fe00 step $200
		fat.read_sector(sec)
		fat.write_xmm_ram(iobuff_ptr, i, $200)
		nmi_vector := word[iobuff_ptr][$fd]
		rst_vector := word[iobuff_ptr][$fe]
		irq_vector := word[iobuff_ptr][$ff]
		sec := fat.get_dent_next_sector(@de, sec)
	str(string("Loaded to $C000..$FFFF",13,10))
	str(string("NMI:$"))
	hex(nmi_vector, 4)
	str(string(" RST:$"))
	hex(rst_vector, 4)
	str(string(" IRQ:$"))
	hex(irq_vector, 4)
	crlf

	cpuregs[0] := rst_vector & 255
	cpuregs[1] := (rst_vector >> 8) & 255
	cpuregs[2] := $34
	cpuregs[3] := $00
	cpuregs[4] := $00
	cpuregs[5] := $00
	cpuregs[6] := $ff

#ifdef FILESELECT
	' *********************************************************
	' file selection
	'
	waitcnt(CNT + 2*clkfreq)

	getdir := 1
	repeat y0 from FSEL_TOP - 1 to 15
		gotoxy(0, y0)
		chr(22)
	repeat while getdir
		gotoxy(0, FSEL_TOP - 1)
		str(string("Sorting directory...",22))
		readdir
		row := 0
		topy := 0
		gotoxy(0, FSEL_TOP - 1)
		str(string("Select a file to load ("))
		dec(boty)
		chr("/")
		dec(ndir)
		str(string("):",22))
		speed := 10
		repeat
			repeat y0 from 0 to FSEL_H - 1
				rc := direntry_name(topy + y0, @buff)
				if rc < 0
					buff[0] := " "
					buff[1] := " "
					buff[2] := 0
				gotoxy(0, FSEL_TOP+y0)
				if y0 == row
					str(string("=> "))
				else
					str(string("   "))
				str(@buff)
				if y0 == row
					str(string(" <="))
				chr(22)

			waitcnt(CNT + clkfreq / speed)

			if kbd.keystate($c3)
				if key == $34
					speed := ++speed <# 30
				else
					speed := 10
					key := $34
				if (row < FSEL_H - 1) and (row < boty)
					row++
				elseif topy < boty - (FSEL_H - 1)
					topy++
			elseif key == $c3
				key := 0

			if kbd.keystate($c2)	' Up arrow
				if key == $33
					speed := ++speed <# 30
				else
					speed := 10
					key := $33
				if row > 0
					row--
				elseif topy > 0
					topy--
			elseif key == $c2
				key := 0

			if kbd.keystate($c7)	' Page down
				key := $c7
				if topy + FSEL_H + FSEL_H - 1 < boty
					topy += FSEL_H
				else
					topy := boty - (FSEL_H - 1)
					row := boty - topy

			if kbd.keystate($c6)	' Page up
				key := $c6
				if topy - FSEL_H > 0
					topy -= FSEL_H
				else
					topy := 0
					row := 0

			if kbd.keystate($0d)	' Return
				repeat y0 from 0 to FSEL_H - 1
					gotoxy(0, FSEL_TOP+y0)
					chr(22)
				gotoxy(0, FSEL_TOP)
				getdir := 0
				rc := direntry_name(topy + row, @buff)
				if byte[@buff][0] == "<"
					chdir(topy + row)
					getdir := 1
				quit
		
	' *********************************************************
	' load the selected file
	'
	if byte[@buff] <> 32
		m65_load(@buff)
#endif FILESELECT

#ifdef PSGS
	okay := psg.start(24, 0, @psg0)
	success(string("PSG0"))
	okay := psg.start(25, 0, @psg1)
	success(string("PSG1"))
#endif PSGS

	gfx_switch := 0

	' *************************************************************************************
	' Start the I/O handler cog
	'
	okay := cog := COGNEW(@entry, @io_command) + 1
	if okay =< 0
		failure(string("I/O"),okay)
		return -1
	success(string("I/O"))
	return 0

PUB stop
'' Stop CPU simulator - frees a cog
	if cog
		cogstop(cog~ - 1)

PUB gotoxy(x,y)
	curx := (x #> 0) <# 31
	cury := (y #> 0) <# 15

PUB chr(ch)
	case ch
		22:	' clear to end of line
			bytefill(screen_ptr + cury << 5 + curx, $20, 32 - curx)
		13:	' carriage return
			curx := 0
		12:	' form feed - clear screen
			longfill(screen_ptr, $20202020, 32*16/4)
			curx := 0
			cury := 0
		10:	' line feed
			if ++cury > 15
				longmove(screen_ptr, screen_ptr + 32, 32*15/4)
				longfill(screen_ptr + 32*15, $20202020, 32/4)
				cury := 15
		9:	' tab
			repeat until (curx & 7) == 0
				chr(32)
		8:	' back space
			if curx > 0
				--curx
		other:
			byte[screen_ptr][cury << 5 + curx] := ch
			curx++
			if curx => 32
				chr(13)
				chr(10)

PUB str(src) | i
	repeat i from 0 to strsize(src) - 1
		chr(byte[src][i])

PUB dec(value) | _i
'' Print a decimal number
	result := 0
	if value < 0
		chr("-")
		if (value == NEGX)
			chr("2")
			value += 2_000_000_000
		-value
	_i := 1_000_000_000
	repeat 10
		if value => _i
			chr(value / _i + "0")
			value //= _i
			result~~
		elseif result or _i == 1
			chr("0")
		_i /= 10

PUB crlf
	chr(13)
	chr(10)

PUB dec_size(value)
	if value > 1048576
		dec(value/1048576)
		chr(".")
		dec(10*(value // 1048576) / 1048576)
		str(string("MB"))
	elseif value > 1024
		dec(value/1024)
		chr(".")
		dec(10*(value // 1024) / 1024)
		str(string("KB"))
	else
		dec(value)

PUB hex(value, digits)
'' Print a hexadecimal number

	value <<= (8 - digits) << 2
	repeat digits
		chr(lookupz((value <-= 4) & $F : "0".."9", "A".."F"))

PUB bin(value, digits)
'' Print a binary number

	value <<= 32 - digits
	repeat digits
		chr((value <-= 1) & 1 + "0")

PRI chr_ptr(ch, dst)
	byte[dst++] := ch
	byte[dst] := 0
	return dst

PRI str_ptr(src, dst) | i
	repeat i from 0 to strsize(src) - 1
		byte[dst++] := byte[src][i]
	byte[dst] := 0
	return dst

PRI dec_ptr(value, dst) | _i
'' Print a decimal number
	result := 0
	if value < 0
		dst := chr_ptr("-", dst)
		if (value == NEGX)
			dst := chr_ptr("2", dst)
			value += 2_000_000_000
		-value
	_i := 1_000_000_000
	repeat 10
		if value => _i
			dst := chr_ptr(value / _i + "0", dst)
			value //= _i
			result~~
		elseif result or _i == 1
			dst := chr_ptr("0", dst)
		_i /= 10
	return dst

PRI dec_size_ptr(value, dst)
	if value > 1048576
		dst := dec_ptr(value/1048576, dst)
		dst := chr_ptr(".", dst)
		dst := dec_ptr(10*(value // 1048576) / 1048576, dst)
		dst := str_ptr(string("MB"), dst)
	elseif value > 1024
		dst := dec_ptr(value/1024, dst)
		dst := chr_ptr(".", dst)
		dst := dec_ptr(10*(value // 1024) / 1024, dst)
		dst := str_ptr(string("KB"), dst)
	else
		dst := dec_ptr(value, dst)
	return dst

PUB failure(who, val)
	str(who)
	str(string(" failed: "))
	dec(val)
	crlf

PUB success(who)
	str(who)
	str(string(" initialized, "))
	dec(free_cogs)
	str(string(" cogs free.", 13, 10))

PUB free_cogs : free | i, cogs[8], jmp_0
	jmp_0 := %010111_0001_1111_000000000_000000000
	repeat while (cogs[free] := cognew(@jmp_0, 0)) => 0 
		free++
	if free
		repeat i from 0 to free - 1
			cogstop(cogs[i])
	return free

PRI strncmp(dst, src, n) | i
	repeat i from 0 to n - 1
		if byte[dst][i] <> byte[src][i]
			return byte[dst][i] - byte[src][i]
	return 0

PRI readdir | rec, i, swapped
	rec := fat.opendir(@de)
	boty := 0
	ndir := 0
	repeat while rec => 0
		ndir++
		if byte[@de][fat#dent_name] <> $e5	' deleted entry?
			if (fat.match(@de, string("*.M65")) or fat.isdir(@de)) and not fat.isdot(@de)
				fat.write_xmm_ram(@de, SCRATCH_RAM + boty * 32, 32)
				boty++
		rec := fat.readdir(rec, @de)
	' do a bubble sort of the filenames in memory
	swapped := 1
	repeat until swapped == 0
		swapped := 0
		repeat i from 0 to boty - 2
			' read two directory entries
			fat.read_xmm_ram(@buff, SCRATCH_RAM + i * fat#dent_bytes, 2*fat#dent_bytes)
			' and compare the names
			if strncmp(@buff, @buff + fat#dent_bytes, 8) > 0
				fat.write_xmm_ram(@buff, SCRATCH_RAM + (i + 1) * fat#dent_bytes, fat#dent_bytes)
				fat.write_xmm_ram(@buff + fat#dent_bytes, SCRATCH_RAM + i * fat#dent_bytes, fat#dent_bytes)
				swapped++

PRI direntry_name(n, buf) | i, temp[fat#dent_bytes/4]
	if n < boty
		fat.read_xmm_ram(@temp, SCRATCH_RAM + n * 32, 32)
		if (byte[@temp][fat#dent_attr] & fat#dent_attr_D) <> 0
			' directory
			byte[buf] := "<"
			fat.get_dent_name(@temp, buf + 1)
			i := strsize(buf)
			byte[buf + i + 0] := ">"
			byte[buf + i + 1] := 0
		else
			fat.get_dent_name(@temp, buf)
		return 0
	return -1

PRI chdir(n) | temp[fat#dent_bytes/4]
	if n < boty
		fat.read_xmm_ram(@temp, SCRATCH_RAM + n * fat#dent_bytes, fat#dent_bytes)
		fat.chdir(@temp)
		return 0
	return -1

PRI open_file(fname_ptr) : dent | size, sec
	dent := fat.find_file(fname_ptr)
	if dent < 0
		str(fname_ptr)
		str(string(" not found.",13,10))
		return -1
	bytemove(@de, dent, 32)
	fat.get_dent_name(dent, @name)
	size := fat.get_dent_size(@de)
	sec := fat.get_dent_first_sector(@de)
	fat.get_dent_date_time(@de, @datetime)
	str(@name)
	str(string(9,"sector "))
	dec(sec)
	str(string(9,"size "))
	dec_size(size)
	crlf

PRI m65_load(fname_ptr) | dent, size, sec, nsec, pos, state, badr, blen, val
	dent := open_file(fname_ptr)
	if dent < 0
		return -1
	bytemove(@de, dent, 32)
	fat.get_dent_name(@de, @name)
	size := fat.get_dent_size(@de)
	sec := fat.get_dent_first_sector(@de)

	pos := $200
	state := 0
	badr := 0
	blen := 8192
	if size == CONSTANT(8192+64+7)
		' Old M65 format: 8192 bytes RAM, 64 bytes graphics bits, 7 bytes CPU registers
		repeat
			if pos == $200
				if sec < 0
					quit
				nsec := fat.get_dent_next_sector(@de, sec)
				fat.read_sector(sec)
				sec := nsec
				pos := 0
			case state
				0:	' expecting blen bytes of RAM data
					val := $200 - pos
					val <#= blen
					fat.write_xmm_ram(iobuff_ptr + pos, badr, val)
					badr += val
					pos += val
					blen -= val
					if blen == 0
						' copy video RAM to frame buffer
						fat.read_xmm_ram(screen_ptr, $200, $200)
						badr := 0
						blen := 64
						state++

				1:	' expecting 64 bytes of chunky graphics bits
					if blen-- > 0
						byte[graphics_ptr][badr++] := byte[iobuff_ptr][pos++]
						if blen == 0
							badr := 0
							blen := 7
							state++

				2:	' expecting 7 bytes of CPU registers (PCL, PCH, PSW, A, IX, IY, SP)
					if blen-- > 0
						byte[@cpuregs][badr++] := byte[iobuff_ptr][pos++]
						if blen == 0
							quit
	else
		' New M65 format
		repeat
			if pos == $200
				if sec < 0
					quit
				nsec := fat.get_dent_next_sector(@de, sec)
				fat.read_sector(sec)
				sec := nsec
				pos := 0
			case state
				0:	' expecting LSB of version
					version := byte[iobuff_ptr][pos++]
					state++

				1:	' expecting MSB of version
					version := 256 * byte[iobuff_ptr][pos++]
					state++

				2:	' expecting LSB of version
					ramsize := byte[iobuff_ptr][pos++]
					state++

				3:	' expecting MSB of version
					ramsize := 256 * byte[iobuff_ptr][pos++]
					badr := 0
					blen := ramsize
					state++

				4:	' expecting blen bytes of RAM data
					val := $200 - pos
					val <#= blen
					fat.write_xmm_ram(iobuff_ptr + pos, badr, val)
					badr += val
					pos += val
					blen -= val
					if blen == 0
						fat.read_xmm_ram(screen_ptr, $200, $400)
						badr := $bfc0
						blen := 16
						state++

				5:	' expecting 16 bytes of 1st 6522 data ($bfc0..$bfcf)
					if blen-- > 0
						fat.write_xmm_ram(iobuff_ptr + pos, badr, 1)
						pos++
						badr++
					else
						badr := $bfd0
						blen := 16
						state++

				6:	' expecting 16 bytes of 2nd 6522 data ($bfc0..$bfcf)
					if blen-- > 0
						fat.write_xmm_ram(iobuff_ptr + pos, badr, 1)
						pos++
						badr++
					else
						badr := $bff0
						blen := 16
						state++

				7:	' expecting 16 bytes of microtan I/O ($bff0..$bfff)
					if blen-- > 0
						fat.write_xmm_ram(iobuff_ptr + pos, badr, 1)
						pos++
						badr++
					else
						badr := $bc04
						blen := 1
						state++

				8:	' expecting 1 byte of invaders sound ($bc04)
					if blen-- > 0
						fat.write_xmm_ram(iobuff_ptr + pos, badr, 1)
						pos++
						badr++
					else
						blen := 1
						state++

				9:	' expecting 1 byte of graphics flag
					if blen-- > 0
						gfx_switch := byte[iobuff_ptr][pos++]
					else
						badr := 0
						blen := 16
						state++

				10:	' expecting 16 bytes of 1st AY-3-8910 registers
					if blen-- > 0
						byte[psg0][badr++] := byte[iobuff_ptr][pos++]
					else
						badr := 0
						blen := 16
						state++

				11:	' expecting 16 bytes of 2nd AY-3-8910 registers
					if blen-- > 0
						byte[psg1][badr++] := byte[iobuff_ptr][pos++]
					else
						badr := 0
						blen := 64
						state++

				12:	' expecting 64 bytes of chunky graphics bits
					if blen-- > 0
						byte[graphics_ptr][badr++] := byte[iobuff_ptr][pos++]
					else
						badr := 0
						blen := 7
						state++

				13:	' expecting 7 bytes of CPU registers (PCL, PCH, PSW, A, IX, IY, SP)
					if blen-- > 0
						byte[@cpuregs][badr++] := byte[iobuff_ptr][pos++]
					if blen == 0
						quit

PUB command_addr
'' Return the address of the command long
	return @io_command

PUB irq_addr
'' Return the address of the irq long
	return @irq

PUB videoram_addr
'' Return the address of the videoram
	return @videoram

PUB graphicsram_addr
'' Return the address of the graphicsram
	return @graphicsram

PUB cpuregs_addr
'' Return the address of the CPU registers
	return @cpuregs

DAT
'---------------------------------------------------------------------------------------------------------

		org	0

'---------------------------------------------------------------------------------------------------------
entry
			wrlong	null, iocmd_ptr
command
			tjz	delayed_nmi, #:no_nmi
			sub	delayed_nmi, #1		wz
		if_nz	jmp	#:no_nmi
			tjz	irq_ptr, #:no_nmi		' don't write, if the pointer is NULL
			wrlong	minus1, irq_ptr
:no_nmi
			rdlong	kbd_head, kbd_head_ptr
			rdlong	kbd_tail, kbd_tail_ptr
			cmp	kbd_head, kbd_tail	wz	' new keypress?
		if_nz	call	#kbd_read

			rdlong	iocmd, iocmd_ptr
			mov	t1, iocmd
			shr	t1, #24			wz
		if_z	jmp	#command

			mov	addr, iocmd			' get command word
			shr	addr, #8			' extract addr
			and	addr, low_16bits		' lower 16 bits

			cmp	t1, #io_cmd_out		wz
		if_z	jmp	#output
			cmp	t1, #io_cmd_in		wz
		if_z	jmp	#input
			cmp	t1, #io_cmd_break	wz
		if_z	jmp	#break
command_0
			and	iocmd, low_24bits
			wrlong	iocmd, iocmd_ptr
			jmp	#command

'**************************************************************************************
'
' Read a key from the keyboard and issue an interrupt
'
kbd_read
			mov	t1, kbd_tail
			shl	t1, #1				' index words
			add	t1, kbd_keys_ptr
			rdword	kbd_data, t1
			add	kbd_tail, #1
			and	kbd_tail, #15
			wrlong	kbd_tail, kbd_tail_ptr

			cmp	kbd_data, kbd_ctrlaltdel	WZ	' ctrl+alt+delete?
		if_z	rdbyte	t1, #4				' get the _clkmode shadow register
		if_z	or	t1, #$80			' set the RESET bit
		if_z	clkset	t1				' reboot
			test	kbd_data, kbd_ctrl		WZ	' is it a control code?
		if_nz	and	kbd_data, #$1f			' make it $00..$1f
			cmp	kbd_data, #$c0		WZ	' cursor left?
		if_z	mov	kbd_data, #("S" & $1f)
			cmp	kbd_data, #$c1		WZ	' cursor right?
		if_z	mov	kbd_data, #("D" & $1f)
			cmp	kbd_data, #$c2		WZ	' cursor up?
		if_z	mov	kbd_data, #("E" & $1f)
			cmp	kbd_data, #$c3		WZ	' cursor down?
		if_z	mov	kbd_data, #("X" & $1f)
			cmp	kbd_data, #$c6		WZ	' page up?
		if_z	mov	kbd_data, #("R" & $1f)
			cmp	kbd_data, #$c7		WZ	' page down?
		if_z	mov	kbd_data, #("C" & $1f)
			cmp	kbd_data, #$c8		WZ	' backspace?
		if_z	mov	kbd_data, #8
			cmp	kbd_data, #$c9		WZ	' delete?
		if_z	mov	kbd_data, #127
			cmp	kbd_data, #$ca		WZ	' insert?
		if_z	mov	kbd_data, #("V" & $1f)
			cmp	kbd_data, #$cb		WZ	' upper code for Esc?
		if_z	mov	kbd_data, #27			' Escape character

			mov	kbd_flag, #$80			' set keyboard flag
			tjz	irq_ptr, #kbd_read_ret		' return if irq_ptr is 0
			mov	t1, #1
			wrlong	t1, irq_ptr			' request an IRQ
kbd_read_ret
			ret		

'**************************************************************************************
'
' Read i/o addr and data to write to
'
output
			mov	data, iocmd			' get command word
			and	data, #$ff			' extract data

			mov	t1, addr
			andn	t1, #$1ff
			cmp	t1, screen_base		wz
		if_z	jmp	#output_scr

			mov	t1, addr
			andn	t1, #3
			cmp	t1, psg_base		wz	' a write to the PSG registers?
		if_z	jmp	#output_psg

			cmp	addr, io_base		wc
		if_c	jmp	#command_0
			and	addr, #3		wz	' mask 4 addresses
		if_z	jmp	#output_0
			cmp	addr, #1		wz
		if_z	jmp	#output_1
			cmp	addr, #2		wz
		if_z	jmp	#output_2
output_3
			mov	gfx_switch, #0
			jmp	#command_0
output_2
			' write keyboard scan pattern (only for the hex keypad)
			jmp	#command_0
output_1
			mov	delayed_nmi, #8
			jmp	#command_0
output_0
			mov	kbd_flag, #0			' reset keyboard flag
			wrlong	null, irq_ptr			' clear IRQ
			jmp	#command_0
output_scr
			and	addr, #$1ff
			mov	t1, addr
			add	addr, screen_ptr
			wrbyte	data, addr
			mov	addr, t1
			shr	addr, #5			' long index
			shl	addr, #2			' * 4
			add	addr, graphics_ptr
			mov	data, #1			' make bitmask
			shl	data, t1			' one of bit 0..31
			rdlong	gfx_data, addr			' get graphics bitmap
			cmp	gfx_switch, #1		wz	' graphics switch on?
			muxz	gfx_data, data			' set or clear graphics bit
			wrlong	gfx_data, addr			' write back graphics bitmap
			jmp	#command_0

output_psg
			and	addr, #3		wz	' which register?
		if_z	jmp	#output_psg0
			cmp	addr, #1		wz
		if_z	jmp	#output_psg1
			cmp	addr, #2		wz
		if_z	jmp	#output_psg2
output_psg3
			mov	t1, psg1_index
			add	t1, psg1_ptr
			cmp	psg1_index, #13		wz
		if_z	and	data, #15
			wrbyte	data, t1
			jmp	#command_0

output_psg2
			mov	psg1_index, data
			jmp	#command_0

output_psg1
			mov	t1, psg0_index
			add	t1, psg0_ptr
			cmp	psg0_index, #13		wz
		if_z	and	data, #15
			wrbyte	data, t1
			jmp	#command_0

output_psg0
			mov	psg0_index, data
			jmp	#command_0

'**************************************************************************************
'
' Read i/o addr to read from
'
input
			andn	iocmd, #$ff			' clear the input data
			mov	t1, addr
			andn	t1, #$1ff
			cmp	t1, screen_base		wz
		if_z	jmp	#input_scr

			mov	t1, addr
			andn	t1, #3
			cmp	t1, psg_base		wz	' a write to the PSG registers?
		if_z	jmp	#input_psg

			cmp	addr, io_base		wc
		if_c	jmp	#input_ff
			and	addr, #%00000011	wz	' mask lower two bits
		if_z	jmp	#input_0
			cmp	addr, #%00000001	wz
		if_z	jmp	#input_1
			cmp	addr, #%00000010	wz
		if_z	jmp	#input_2
input_3
			mov	data, kbd_data
			or	data, kbd_flag
			jmp	#input_data
input_2
			' unused address
			jmp	#input_ff
input_1
			' unused address
			jmp	#input_ff
input_0
			mov	gfx_switch, #1			' enable graphics write
input_ff
			mov	data, #$ff
input_data
			or	iocmd, data			' return $ff for other i/o addresses
			jmp	#command_0

input_scr
			and	addr, #$1ff
			add	addr, screen_ptr		' + screen base
			rdbyte	data, addr			' fetch data
			jmp	#input_data

input_psg
			and	addr, #2		wz	' psg0 or psg1?
		if_z	jmp	#input_psg0
input_psg1
			mov	t1, psg1_index
			add	t1, psg1_ptr
			rdbyte	data, t1
			jmp	#input_data
input_psg0
			mov	t1, psg0_index
			add	t1, psg0_ptr
			rdbyte	data, t1
			jmp	#input_data

'**************************************************************************************
'
' Output a character to the screen
'
conout_chr
			mov	addr, cur_y			' y
			shl	addr, #5			' * 32
			add	addr, cur_x			' + x
			add	addr, screen_ptr		' + screen address

			cmp	data, #8		wz	' backspace?
		if_z	jmp	#:bsp
			cmp	data, #9		wz	' tab?
		if_z	jmp	#:tab
			cmp	data, #10		wz	' line feed?
		if_z	jmp	#:lf
			cmp	data, #12		wz	' clear screen?
		if_z	jmp	#:cls
			cmp	data, #13		wz	' carriage return?
		if_z	jmp	#:cr
:out
			wrbyte	data, addr
			add	cur_x, #1
:wrap
			cmp	cur_x, #32		wz
		if_nz	jmp	#:exit
			mov	cur_x, #0
:lf
			add	cur_y, #1
			cmp	cur_y, #16		wz
		if_nz	jmp	#:exit
			call	#scroll
			mov	cur_y, #15
			jmp	#:exit

:bsp
			cmp	cur_x, #0		wz, wc
		if_be	jmp	#:exit
			sub	cur_x, #1
			jmp	#:exit
:tab
			mov	data, #32
			wrbyte	data, addr
			add	addr, #1
			add	cur_x, #1
			test	cur_x, #7		wz
		if_nz	jmp	#:tab
			jmp	#:wrap
:cr
			mov	cur_x, #0
			jmp	#:exit
:cls
			mov	addr, screen_ptr
			mov	cur_y, #32*16/4
:loop
			wrlong	x20202020, addr
			add	addr, #4
			djnz	cur_y, #:loop
			mov	cur_x, #0
:exit
conout_chr_ret
			ret		

'**************************************************************************************
'
' Scroll the screen one line up
'
scroll
			mov	addr, screen_ptr
			mov	cur_y, #15
:row
			mov	cur_x, #32/4
:col
			add	addr, #32
			rdlong	data, addr
			sub	addr, #32
			wrlong	data, addr
			add	addr, #4
			djnz	cur_x, #:col
			djnz	cur_y, #:row
			mov	cur_x, #32/4
:fill
			wrlong	x20202020, addr
			add	addr, #4
			djnz	cur_x, #:fill
scroll_ret
			ret


'**************************************************************************************
'
' Output a byte in hexval to the console
'
conout_hex_b
			mov	hexcnt, #2
			shl	hexval, #24
			jmp	#conout_hex

'**************************************************************************************
'
' Output a word in hexval to the console
'
conout_hex_w
			mov	hexcnt, #4
			shl	hexval, #16
			jmp	#conout_hex

'**************************************************************************************
'
' Output a long in hexval to the console
'
conout_hex_l
			mov	hexcnt, #8
conout_hex
			mov	data, hexval
			shr	data, #28
			cmp	data, #10	wz, wc
		if_b	add	data, #48
		if_ae	add	data, #55
			call	#conout_chr
			shl	hexval, #4
			djnz	hexcnt, #conout_hex
conout_hex_ret
conout_hex_l_ret
conout_hex_w_ret
conout_hex_b_ret
			ret

'**************************************************************************************
'
' Output a space to the console
'
conout_space
			mov	data, #32
			call	#conout_chr
conout_space_ret
			ret

'**************************************************************************************
'
' Output a carriage return and line feed to the console
'
conout_crlf
			mov	data, #13
			call	#conout_chr
			mov	data, #10
			call	#conout_chr
conout_crlf_ret
			ret

'**************************************************************************************
'
' Output a string at strptr to the console
'
conout_string
			rdbyte	data, strptr	wz
			add	strptr, #1
		if_z	jmp	#conout_string_ret
			call	#conout_chr
			jmp	#conout_string
conout_string_ret
			ret

'**************************************************************************************
'
' Break - issued when the q6502 encounters an illegal instruction
'
break
			mov	t1, iocmd			' read the address of the register file
			shr	t1, #8				' bits 23..8 to 15..0
			and	t1, low_16bits			' mask only bits 15..0
			mov	strptr, msg_break_ptr	

			call	#conout_string
			rdword	hexval, t1			' read PC
			call	#conout_hex_w
			add	t1, #2
	
			mov	t2, #5
:loop_1			call	#conout_string
			rdbyte	hexval, t1			' read P, A, X, Y, S
			call	#conout_hex_b
			add	t1, #1
			djnz	t2, #:loop_1

			rdbyte	t2, t1
			add	t2, op_len_ptr
			rdbyte	t2, t2
:loop_2			call	#conout_space
			rdbyte	hexval, t1
			call	#conout_hex_b
			add	t1, #1
			djnz	t2, #:loop_2

			call	#conout_crlf

			jmp	#command_0

iocmd_ptr		long	0-0
screen_ptr		long	0-0
graphics_ptr		long	0-0
irq_ptr			long	0-0
cur_x			long	0
cur_y			long	0
iobuff_ptr		long	0-0
x20202020		long	$20202020
x10_000			long	10_000

gfx_switch		long	0
gfx_data		long	0

kbd_head_ptr		long	0-0
kbd_tail_ptr		long	0-0
kbd_keys_ptr		long	0-0
kbd_ctrlaltdel		long	$6c9
kbd_ctrl		long	$200
kbd_data		long	0
kbd_flag		long	0
	
minus1			long	$ffffffff
low_24bits		long	$00ffffff
low_16bits		long	$0000ffff
screen_base		long	$0200
msg_break_ptr		long	@@@msg_break
op_len_ptr		long	@@@op_len_table
psg0_ptr		long	@@@psg0
psg1_ptr		long	@@@psg1
psg_base		long	$bc00
io_base			long	$bfc0					' $bfc0..$bfff is memory mapped i/o
delayed_nmi		long	0

kbd_head		res	1
kbd_tail		res	1

psg0_index		res	1
psg1_index		res	1

iocmd			res	1
addr			res	1
data			res	1
null			res	1

t1			res	1
t2			res	1
t3			res	1
t4			res	1

strptr			res	1
hexval			res	1
hexcnt			res	1

			fit	$1f0
io_command		long	0
irq			long	0
cpuregs			byte	0[8]

videoram		byte	0[32*16]
graphicsram		long	0[16]
cursor			byte	0[8]
sync			long	0

psg0			byte	0[16]		' 1st AY-3-8910
psg1			byte	0[16]		' 2nd AY-3-8910

kbd_rows_table
			byte	0,0,0,0,0,0,0,0

msg_break		byte	"PC:",0
			byte	" P:",0
			byte	" A:",0
			byte	" X:",0
			byte	" Y:",0
			byte	" S:",0
op_len_table
		{00}	byte	1	'' 7 BRK
		{01}	byte	2	'' 6 ORA IDX
		{02}	byte	1	'' 1 ILL
		{03}	byte	2	'' 7 SLO IDX
		{04}	byte	2	'' 3 NOP ZPG
		{05}	byte	2	'' 3 ORA ZPG
		{06}	byte	2	'' 5 ASL ZPG
		{07}	byte	2	'' 5 SLO ZPG
		{08}	byte	1	'' 3 PHP
		{09}	byte	2	'' 2 ORA IMM
		{0a}	byte	1	'' 2 ASL A
		{0b}	byte	2	'' 2 ANC IMM
		{0c}	byte	3	'' 4 NOP ABS
		{0d}	byte	3	'' 4 ORA ABS
		{0e}	byte	3	'' 6 ASL ABS
		{0f}	byte	3	'' 6 SLO ABS
		{10}	byte	2	'' 2-4 BPL REL
		{11}	byte	2	'' 5 ORA IDY page penalty
		{02}	byte	1	'' 1 ILL
		{13}	byte	2	'' 7 SLO IDY
		{14}	byte	2	'' 4 NOP ZPX
		{15}	byte	2	'' 4 ORA ZPX
		{16}	byte	2	'' 6 ASL ZPX
		{17}	byte	2	'' 6 SLO ZPX
		{18}	byte	1	'' 2 CLC
		{19}	byte	3	'' 4 ORA ABY page penalty
		{1a}	byte	1	'' 2 NOP
		{1b}	byte	3	'' 7 SLO ABY
		{1c}	byte	3	'' 4 NOP ABX page penalty
		{1d}	byte	3	'' 4 ORA ABX page penalty
		{1e}	byte	3	'' 7 ASL ABX
		{1f}	byte	3	'' 7 SLO ABX
		{20}	byte	3	'' 6 JSR
		{21}	byte	2	'' 6 AND IDX
		{02}	byte	1	'' 1 ILL
		{23}	byte	2	'' 7 RLA IDX
		{24}	byte	2	'' 3 BIT ZPG
		{25}	byte	2	'' 3 AND ZPG
		{26}	byte	2	'' 5 ROL ZPG
		{27}	byte	2	'' 5 RLA ZPG
		{28}	byte	1	'' 4 PLP
		{29}	byte	2	'' 2 AND IMM
		{2a}	byte	1	'' 2 ROL A
		{2b}	byte	2	'' 2 ANC IMM
		{2c}	byte	3	'' 4 BIT ABS
		{2d}	byte	3	'' 4 AND ABS
		{2e}	byte	3	'' 6 ROL ABS
		{2f}	byte	3	'' 6 RLA ABS
		{30}	byte	2	'' 2-4 BMI REL
		{31}	byte	2	'' 5 AND IDY page penalty
		{02}	byte	1	'' 1 ILL
		{33}	byte	2	'' 7 RLA IDY
		{34}	byte	2	'' 4 NOP ZPX
		{35}	byte	2	'' 4 AND ZPX
		{36}	byte	2	'' 6 ROL ZPX
		{37}	byte	2	'' 6 RLA ZPX
		{38}	byte	1	'' 2 SEC
		{39}	byte	3	'' 4 AND ABY page penalty
		{3a}	byte	1	'' 2 NOP
		{3b}	byte	3	'' 7 RLA ABY
		{3c}	byte	3	'' 4 NOP ABX page penalty
		{3d}	byte	3	'' 4 AND ABX page penalty
		{3e}	byte	3	'' 7 ROL ABX
		{3f}	byte	3	'' 7 RLA ABX
		{40}	byte	1	'' 6 RTI
		{41}	byte	2	'' 6 EOR IDX
		{02}	byte	1	'' 1 ILL
		{43}	byte	2	'' 7 SRE IDX
		{44}	byte	2	'' 3 NOP ZPG
		{45}	byte	2	'' 3 EOR ZPG
		{46}	byte	2	'' 5 LSR ZPG
		{47}	byte	2	'' 5 SRE ZPG
		{48}	byte	1	'' 3 PHA
		{49}	byte	2	'' 2 EOR IMM
		{4a}	byte	1	'' 2 LSR A
		{4b}	byte	2	'' 2 ASR IMM
		{4c}	byte	3	'' 3 JMP ABS
		{4d}	byte	3	'' 4 EOR ABS
		{4e}	byte	3	'' 6 LSR ABS
		{4f}	byte	3	'' 6 SRE ABS
		{50}	byte	2	'' 2-4 BVC REL
		{51}	byte	2	'' 5 EOR IDY page penalty
		{52}	byte	1	'' 1 ILL
		{53}	byte	2	'' 7 SRE IDY
		{54}	byte	2	'' 4 NOP ZPX
		{55}	byte	2	'' 4 EOR ZPX
		{56}	byte	2	'' 6 LSR ZPX
		{57}	byte	2	'' 6 SRE ZPX
		{58}	byte	1	'' 2 CLI
		{59}	byte	3	'' 4 EOR ABY page penalty
		{5a}	byte	1	'' 2 NOP
		{5b}	byte	3	'' 7 SRE ABY
		{5c}	byte	3	'' 4 NOP ABX page penalty
		{5d}	byte	3	'' 4 EOR ABX page penalty
		{5e}	byte	3	'' 7 LSR ABX
		{5f}	byte	3	'' 7 SRE ABX
		{60}	byte	1	'' 6 RTS
		{61}	byte	2	'' 6 ADC IDX
		{62}	byte	1	'' 1 ILL
		{63}	byte	2	'' 7 RRA IDX
		{64}	byte	2	'' 3 NOP ZPG
		{65}	byte	2	'' 3 ADC ZPG
		{66}	byte	2	'' 5 ROR ZPG
		{67}	byte	2	'' 5 RRA ZPG
		{68}	byte	1	'' 4 PLA
		{69}	byte	2	'' 2 ADC IMM
		{6a}	byte	1	'' 2 ROR A
		{6b}	byte	2	'' 2 ARR IMM
		{6c}	byte	3	'' 5 JMP IND
		{6d}	byte	3	'' 4 ADC ABS
		{6e}	byte	3	'' 6 ROR ABS
		{6f}	byte	3	'' 6 RRA ABS
		{70}	byte	2	'' 2-4 BVS REL
		{71}	byte	2	'' 5 ADC IDY page penalty
		{72}	byte	1	'' 1 ILL
		{73}	byte	2	'' 7 RRA IDY
		{74}	byte	2	'' 4 NOP ZPX
		{75}	byte	2	'' 4 ADC ZPX
		{76}	byte	2	'' 6 ROR ZPX
		{77}	byte	2	'' 6 RRA ZPX
		{78}	byte	1	'' 2 SEI
		{79}	byte	3	'' 4 ADC ABY page penalty
		{7a}	byte	1	'' 2 NOP
		{7b}	byte	3	'' 7 RRA ABY
		{7c}	byte	3	'' 4 NOP ABX page penalty
		{7d}	byte	3	'' 4 ADC ABX page penalty
		{7e}	byte	3	'' 7 ROR ABX
		{7f}	byte	3	'' 7 RRA ABX
		{80}	byte	2	'' 2 NOP IMM
		{81}	byte	2	'' 6 STA IDX
		{82}	byte	2	'' 2 NOP IMM
		{83}	byte	2	'' 6 SAX IDX
		{84}	byte	2	'' 3 STY ZPG
		{85}	byte	2	'' 3 STA ZPG
		{86}	byte	2	'' 3 STX ZPG
		{87}	byte	2	'' 3 SAX ZPG
		{88}	byte	1	'' 2 DEY
		{89}	byte	2	'' 2 NOP IMM
		{8a}	byte	1	'' 2 TXA
		{8b}	byte	2	'' 2 AXA IMM
		{8c}	byte	3	'' 4 STY ABS
		{8d}	byte	3	'' 4 STA ABS
		{8e}	byte	3	'' 4 STX ABS
		{8f}	byte	3	'' 4 SAX ABS
		{90}	byte	2	'' 2-4 BCC REL
		{91}	byte	2	'' 6 STA IDY
		{92}	byte	1	'' 1 ILL
		{93}	byte	2	'' 5 SAH IDY
		{94}	byte	2	'' 4 STY ZPX
		{95}	byte	2	'' 4 STA ZPX
		{96}	byte	2	'' 4 STX ZPY
		{97}	byte	2	'' 4 SAX ZPY
		{98}	byte	1	'' 2 TYA
		{99}	byte	3	'' 5 STA ABY
		{9a}	byte	1	'' 2 TXS
		{9b}	byte	3	'' 5 SSH ABY
		{9c}	byte	3	'' 5 SYH ABX
		{9d}	byte	3	'' 5 STA ABX
		{9e}	byte	3	'' 5 SXH ABY
		{9f}	byte	3	'' 5 SAH ABY
		{a0}	byte	2	'' 2 LDY IMM
		{a1}	byte	2	'' 6 LDA IDX
		{a2}	byte	2	'' 2 LDX IMM
		{a3}	byte	2	'' 6 LAX IDX
		{a4}	byte	2	'' 3 LDY ZPG
		{a5}	byte	2	'' 3 LDA ZPG
		{a6}	byte	2	'' 3 LDX ZPG
		{a7}	byte	2	'' 3 LAX ZPG
		{a8}	byte	1	'' 2 TAY
		{a9}	byte	2	'' 2 LDA IMM
		{aa}	byte	1	'' 2 TAX
		{ab}	byte	2	'' 2 OAL IMM
		{ac}	byte	3	'' 4 LDY ABS
		{ad}	byte	3	'' 4 LDA ABS
		{ae}	byte	3	'' 4 LDX ABS
		{af}	byte	3	'' 4 LAX ABS
		{b0}	byte	2	'' 2-4 BCS REL
		{b1}	byte	2	'' 5 LDA IDY page penalty
		{b2}	byte	1	'' 1 ILL
		{b3}	byte	2	'' 5 LAX IDY page penalty
		{b4}	byte	2	'' 4 LDY ZPX
		{b5}	byte	2	'' 4 LDA ZPX
		{b6}	byte	2	'' 4 LDX ZPY
		{b7}	byte	2	'' 4 LAX ZPY
		{b8}	byte	1	'' 2 CLV
		{b9}	byte	3	'' 4 LDA ABY page penalty
		{ba}	byte	1	'' 2 TSX
		{bb}	byte	3	'' 4 AST ABY page penalty
		{bc}	byte	3	'' 4 LDY ABX page penalty
		{bd}	byte	3	'' 4 LDA ABX page penalty
		{be}	byte	3	'' 4 LDX ABY page penalty
		{bf}	byte	3	'' 4 LAX ABY page penalty
		{c0}	byte	2	'' 2 CPY IMM
		{c1}	byte	2	'' 6 CMP IDX
		{c2}	byte	2	'' 2 NOP IMM
		{c3}	byte	2	'' 7 DCP IDX
		{c4}	byte	2	'' 3 CPY ZPG
		{c5}	byte	2	'' 3 CMP ZPG
		{c6}	byte	2	'' 5 DEC ZPG
		{c7}	byte	2	'' 5 DCP ZPG
		{c8}	byte	1	'' 2 INY
		{c9}	byte	2	'' 2 CMP IMM
		{ca}	byte	1	'' 2 DEX
		{cb}	byte	2	'' 2 ASX IMM
		{cc}	byte	3	'' 4 CPY ABS
		{cd}	byte	3	'' 4 CMP ABS
		{ce}	byte	3	'' 6 DEC ABS
		{cf}	byte	3	'' 6 DCP ABS
		{d0}	byte	2	'' 2-4 BNE REL
		{d1}	byte	2	'' 5 CMP IDY page penalty
		{d2}	byte	1	'' 1 ILL
		{d3}	byte	2	'' 7 DCP IDY
		{d4}	byte	2	'' 4 NOP ZPX
		{d5}	byte	2	'' 4 CMP ZPX
		{d6}	byte	2	'' 6 DEC ZPX
		{d7}	byte	2	'' 6 DCP ZPX
		{d8}	byte	1	'' 2 CLD
		{d9}	byte	3	'' 4 CMP ABY page penalty
		{da}	byte	1	'' 2 NOP
		{db}	byte	3	'' 7 DCP ABY
		{dc}	byte	3	'' 4 NOP ABX page penalty
		{dd}	byte	3	'' 4 CMP ABX page penalty
		{de}	byte	3	'' 7 DEC ABX
		{df}	byte	3	'' 7 DCP ABX
		{e0}	byte	2	'' 2 CPX IMM
		{e1}	byte	2	'' 6 SBC IDX
		{e2}	byte	2	'' 2 NOP IMM
		{e3}	byte	2	'' 7 ISB IDX
		{e4}	byte	2	'' 3 CPX ZPG
		{e5}	byte	2	'' 3 SBC ZPG
		{e6}	byte	2	'' 5 INC ZPG
		{e7}	byte	2	'' 5 ISB ZPG
		{e8}	byte	1	'' 2 INX
		{e9}	byte	2	'' 2 SBC IMM
		{ea}	byte	1	'' 2 NOP
		{eb}	byte	2	'' 2 SBC IMM
		{ec}	byte	3	'' 4 CPX ABS
		{ed}	byte	3	'' 4 SBC ABS
		{ee}	byte	3	'' 6 INC ABS
		{ef}	byte	3	'' 6 ISB ABS
		{f0}	byte	2	'' 2-4 BEQ REL
		{f1}	byte	2	'' 5 SBC IDY page penalty
		{f2}	byte	1	'' 1 ILL
		{f3}	byte	2	'' 7 ISB IDY
		{f4}	byte	2	'' 4 NOP ZPX
		{f5}	byte	2	'' 4 SBC ZPX
		{f6}	byte	2	'' 6 INC ZPX
		{f7}	byte	2	'' 6 ISB ZPX
		{f8}	byte	1	'' 2 SED
		{f9}	byte	3	'' 4 SBC ABY page penalty
		{fa}	byte	1	'' 2 NOP
		{fb}	byte	3	'' 7 ISB ABY
		{fc}	byte	3	'' 4 NOP ABX page penalty
		{fd}	byte	3	'' 4 SBC ABX page penalty
		{fe}	byte	3	'' 7 INC ABX
		{ff}	byte	3	'' 7 ISB ABX

''      MIT LICENSE
{{
'  Permission is hereby granted, free of charge, to any person obtaining
'  a copy of this software and associated documentation files
'  (the "Software"), to deal in the Software without restriction,
'  including without limitation the rights to use, copy, modify, merge,
'  publish, distribute, sublicense, and/or sell copies of the Software,
'  and to permit persons to whom the Software is furnished to do so,
'  subject to the following conditions:
'
'  The above copyright notice and this permission notice shall be included
'  in all copies or substantial portions of the Software.
'
'  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
'  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
'  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
'  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
'  CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
'  TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
'  SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
}}
