'******************************************************************************
'*  microtan.spin - Microtan emulation
'*
'*  (c) Juergen Buchmueller <pullmoll@t-online.de>
'*
'* $Id: microtan.spin,v 1.2 2010-05-27 19:13:03 pm Exp $
'******************************************************************************
CON
	_xinfreq		= 5_000_000		' Quartz is 5MHz
	_clkmode		= xtal1 + pll16x	' System clock is 80MHz

' We need a Z80 to run
#define CPU_M6502

' Memory access dispatch tables for Microtan in q6502
#define MICROTAN

' Memory pages enabled for Microtan
#define PAGED_MEMORY

' If you want to use 1pin TV out, define TV and the video norm
'#define TV
'#define NTSC
#define PAL

' If you want to use VGA out, define VGA and disable the #define TV above
#define VGA

'#define TriBladeProp
'#define RamBladeProp
#define DracBladeProp

' Uncomment to emulate interrupts (NMI and IRQ)
#define INTERRUPTS

#ifdef TriBladeProp
' serial I/O pins on the TriBladeProp #2
	SI			= 31
	SO			= 30
' SD pins on the TriBladeProp #2
	doPin			= 9
	clkPin			= 28
	diPin			= 8
	csPin			= 14
' TV out pin
	tvPin			= 14
' Keyboard pins
	kbDpin			= 26
	kbCpin			= 27
' audio pin
	auPin			= 16	' ???
#endif

#ifdef RamBladeProp
' serial I/O pins on the RamBladeProp
	SI			= 23
	SO			= 22
' microSD pins on RamBladeProp
	doPin			= 24
	clkPin			= 26
	diPin			= 25
	csPin			= 19
' TV out pin
	tvPin			= 14
' Keyboard pins
	kbDpin			= 26
	kbCpin			= 27
' audio pin
	auPin			= 22	' ??? instead of SO?
#endif

#ifdef DracBladeProp
' serial I/O pins on the DracBladeProp
	SI			= 31
	SO			= 30
' SD pins on the DracBladeProp
	doPin			= 12
	clkPin			= 13
	diPin			= 14
	csPin			= 15
' TV out pin
	tvPin			= 16
' Keyboard pins
	kbDpin			= 26
	kbCpin			= 27
' audio pin
	auPin			= 24	' SIO2 TX
#endif

VAR
	byte	pins[10]
	long	params[4]

OBJ
	io	: "io"
	q6502	: "q6502"

PUB     main : okay
	pins[ 0] := SI					' currently unused
	pins[ 1] := SO
	pins[ 2] := doPin
	pins[ 3] := clkPin
	pins[ 4] := diPin
	pins[ 5] := csPin
	pins[ 6] := kbDpin
	pins[ 7] := kbCpin
	pins[ 8] := auPin
	pins[ 9] := tvPin

	okay := io.start(@pins)
	if okay < 0
		repeat

	params[0] := io.command_addr
	params[1] := io.irq_addr
	params[2] := io.videoram_addr
	params[3] := io.graphicsram_addr
	okay := q6502.start(@params, io.cpuregs_addr)
	if okay < 0
		io.failure(string("Q6502"),okay)

''      MIT LICENSE
{{
'  Permission is hereby granted, free of charge, to any person obtaining
'  a copy of this software and associated documentation files
'  (the "Software"), to deal in the Software without restriction,
'  including without limitation the rights to use, copy, modify, merge,
'  publish, distribute, sublicense, and/or sell copies of the Software,
'  and to permit persons to whom the Software is furnished to do so,
'  subject to the following conditions:
'
'  The above copyright notice and this permission notice shall be included
'  in all copies or substantial portions of the Software.
'
'  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
'  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
'  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
'  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
'  CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
'  TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
'  SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
}}
