CON
	VIA_PB		=	0	' port B
	VIA_PA		=	1	' port A
	VIA_DDRB	=	2	' data direction register port B
	VIA_DDRA	=	3	' data direction register port A
	VIA_T1CL	=	4	' timer 1 counter lo
	VIA_T1CH	=	5	' timer 1 counter hi
	VIA_T1LL	=	6	' timer 1 latch lo
	VIA_T1LH	=	7	' timer 1 latch hi
	VIA_T2CL	=	8	' timer 2 counter lo
	VIA_T2CH	=	9	' timer 2 counter hi
	VIA_SR		=	10	' shift register
	VIA_ACR		=	11	' async control register
	VIA_PCR		=	12	' port control register
	VIA_IFR		=	13	' interrupt flag register
	VIA_IER		=	14	' interrupt enable register
	VIA_PANH	=	15	' port A - no handshake

'*******************************************
'
' PCR bits
'
' CA1_LOW_TO_HIGH(c)		(c & $01)
' CA1_HIGH_TO_LOW(c)		(!(c & $01))
'
' CB1_LOW_TO_HIGH(c)		(c & $10)
' CB1_HIGH_TO_LOW(c)		(!(c & $10))
'
' CA2_INPUT(c)			(!(c & $08))
' CA2_LOW_TO_HIGH(c)		((c & $0c) == $04)
' CA2_HIGH_TO_LOW(c)		((c & $0c) == $00)
' CA2_IND_IRQ(c)		((c & $0a) == $02)
'
' CA2_OUTPUT(c)			(c & $08)
' CA2_AUTO_HS(c)		((c & $0c) == $08)
' CA2_HS_OUTPUT(c)		((c & $0e) == $08)
' CA2_PULSE_OUTPUT(c)		((c & $0e) == $0a)
' CA2_FIX_OUTPUT(c)		((c & $0c) == $0c)
' CA2_OUTPUT_LEVEL(c)		((c & $02) >> 1)
'
' CB2_INPUT(c)			(!(c & $80))
' CB2_LOW_TO_HIGH(c)		((c & $c0) == $40)
' CB2_HIGH_TO_LOW(c)		((c & $c0) == $00)
' CB2_IND_IRQ(c)		((c & $a0) == $20)
'
' CB2_OUTPUT(c)			(c & $80)
' CB2_AUTO_HS(c)		((c & $c0) == $80)
' CB2_HS_OUTPUT(c)		((c & $e0) == $80)
' CB2_PULSE_OUTPUT(c)		((c & $e0) == $a0)
' CB2_FIX_OUTPUT(c)		((c & $c0) == $c0)
' CB2_OUTPUT_LEVEL(c)		((c & $20) >> 5)
'
'
' ACR bits
'
' PA_LATCH_ENABLE(c)		(c & $01)
' PB_LATCH_ENABLE(c)		(c & $02)
'
' SR_DISABLED(c)		(!(c & $1c))
' SI_T2_CONTROL(c)		((c & $1c) == $04)
' SI_O2_CONTROL(c)		((c & $1c) == $08)
' SI_EXT_CONTROL(c)		((c & $1c) == $0c)
' SO_T2_RATE(c)			((c & $1c) == $10)
' SO_T2_CONTROL(c)		((c & $1c) == $14)
' SO_O2_CONTROL(c)		((c & $1c) == $18)
' SO_EXT_CONTROL(c)		((c & $1c) == $1c)
'
' T1_SET_PB7(c)			(c & $80)
' T1_CONTINUOUS(c)		(c & $40)
' T2_COUNT_PB6(c)		(c & $20)
'
' Interrupt flags
' INT_CA2	$01
' INT_CA1	$02
' INT_SR	$04
' INT_CB2	$08
' INT_CB1	$10
' INT_T2	$20
' INT_T1	$40
' INT_ANY	$80

DAT
			org	0

write_pb
			test	acr, #%10000000		wz	' T1_SET_PB7?
		if_nz	and	data, #$7f
		if_nz	and	out_b, #$80
		if_nz	or	out_b, data
		if_z	mov	out_b, data
			test	ddr_b, #$ff		wz	' any output bits?
		if_nz	mov	wr_b, out_b
		if_nz	and	wr_b, ddr_b
		if_nz	mov	data, ddr_b
		if_nz	xor	data, #$ff
		if_nz	or	wr_b, data

via_read_regs
			movd	:read, #pb
			mov	count, #16
:read			rdbyte	0-0, via_ptr
			add	via_ptr, #1
			djnz	count, #:read
			sub	via_ptr, #16
via_read_regs_ret
			ret
